"""bed2gff.py - convert bed to gff/gtf
===================================

:Tags: Genomics Intervals BED GFF Conversion

Purpose
-------

This script converts a :term:`bed`-formatted file to a :term:`gff` or
:term:`gtf`-formatted file.

It aims to populate the appropriate fields in the :term:`gff` file
with columns in the :term:`bed` file.

If ``--as-gtf`` is set and a name column in the :term:`bed` file is
present, its contents will be set as ``gene_id`` and
``transcript_id``. Otherwise, a numeric ``gene_id`` or
``transcript_id`` will be set according to ``--id-format``.

Usage
-----

Example::

   # Preview input bed file
   zcat tests/bed2gff.py/bed3/bed.gz | head
   # Convert BED to GFF format
   cgat bed2gff.py < tests/bed2gff.py/bed3/bed.gz > test1.gff
   # View converted file (excluding logging information)
   cat test1.gtf | grep -v "#" | head


+------+-----+------+-------+-------+---+---+---+---------------------------------------+
|chr1  |bed  |exon  |501    |1000   |.  |.  |.  |gene_id "None"; transcript_id "None";  |
+------+-----+------+-------+-------+---+---+---+---------------------------------------+
|chr1  |bed  |exon  |15001  |16000  |.  |.  |.  |gene_id "None"; transcript_id "None";  |
+------+-----+------+-------+-------+---+---+---+---------------------------------------+

Example::

   # Convert BED to GTF format
   cgat bed2gff.py --as-gtf < tests/bed2gff.py/bed3/bed.gz > test2.gtf
   # View converted file (excluding logging information)
   cat test2.gtf | grep -v "#" | head

+------+-----+------+-------+-------+---+---+---+-----------------------------------------------+
|chr1  |bed  |exon  |501    |1000   |.  |.  |.  |gene_id "00000001"; transcript_id "00000001";  |
+------+-----+------+-------+-------+---+---+---+-----------------------------------------------+
|chr1  |bed  |exon  |15001  |16000  |.  |.  |.  |gene_id "00000002"; transcript_id "00000002";  |
+------+-----+------+-------+-------+---+---+---+-----------------------------------------------+

Type::

   cgat bed2gff.py --help

for command line help.

Command line options
--------------------

"""
import sys
import CGAT.Experiment as E
import CGAT.GTF as GTF
import CGAT.Bed as Bed


def main(argv=sys.argv):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-a", "--as-gtf", dest="as_gtf", action="store_true",
                      help="output as gtf.")

    parser.add_option(
        "-f", "--id-format", dest="id_format", type="string",
        help="format for numeric identifier if --as-gtf is set and "
        "no name in bed file [%default].")

    parser.set_defaults(as_gtf=False,
                        id_format="%08i",
                        test=None)

    (options, args) = E.Start(parser, add_pipe_options=True)

    as_gtf = options.as_gtf
    id_format = options.id_format

    if as_gtf:
        gff = GTF.Entry()
    else:
        gff = GTF.Entry()

    gff.source = "bed"
    gff.feature = "exon"

    ninput, noutput, nskipped = 0, 0, 0

    id = 0
    for bed in Bed.iterator(options.stdin):

        ninput += 1

        gff.contig = bed.contig
        gff.start = bed.start
        gff.end = bed.end
        if bed.fields and len(bed.fields) >= 3:
            gff.strand = bed.fields[2]
        else:
            gff.strand = "."

        if bed.fields and len(bed.fields) >= 2:
            gff.score = bed.fields[1]

        if as_gtf:
            if bed.fields:
                gff.gene_id = bed.fields[0]
                gff.transcript_id = bed.fields[0]
            else:
                id += 1
                gff.gene_id = id_format % id
                gff.transcript_id = id_format % id
        else:
            if bed.fields:
                gff.source = bed.fields[0]

        options.stdout.write(str(gff) + "\n")

        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Genomics.py - Tools for working with genomic data
=================================================

:Tags: Python

Reference
---------

"""

import numpy
import os
import string
import re
import hashlib
import base64
import tempfile
import sys
from functools import reduce

try:
    import alignlib_lite
except ImportError:
    pass

from CGAT import AString as AString

global_last_filename_genome = None
global_forward_sequences = {}
global_last_sbjct_token = None

if sys.version_info.major >= 3:
    global_translator = str.maketrans("ACGTacgt", "TGCAtgca")
else:
    global_translator = string.maketrans("ACGTacgt", "TGCAtgca")


def complement(s):
    """reverse complement a sequence.

    >>> complement("ACATACATACTA")
    'TAGTATGTATGT'

    Returns
    -------
    string
    """
    return s[::-1].translate(global_translator)


def GetHID(sequence):
    """returns a hash value for a sequence.

    The hash value is computed using md5 and converted
    into printable characters.

    >>> GetHID("ACATACATACTA")
    'trcPGx9VNT36XMlG0XvUBQ'

    Returns
    -------
        A hash value
    """

    # do the encryption
    h = hashlib.md5(sequence).digest()

    # map to printable letters: hid has length 22, so the padded '='
    # are truncated.
    r = base64.encodestring(h)[0:22]

    # finally substitute some characters:
    # '/' for '_', so we have legal file names
    # '[' for '+' and ']' for '=' for internet-applications

    hid = string.replace(r, '/', '_')
    hid = string.replace(hid, '+', '[')
    hid = string.replace(hid, '=', ']')

    return hid


def String2Location(s):
    """convert a string to location information.

    >>> String2Location("chr1:12:15")
    ('chr1', '+', 12, 15)

    Returns
    -------
    contig : string
    strand : string
    start : int
    end : int
    """

    data = re.split("[:]+", s)
    if len(data) == 3:
        return data[0], "+", int(data[1]), int(data[2])
    elif len(data) == 4:
        return data[0], data[1], int(data[2]), int(data[3])
    else:
        raise ValueError("unknown format %s" % (s))


def readContigSizes(infile):
    """read sizes of contigs from file.

    Arguments
    ---------
    infile : string
        Filename of :term:`tsv` separated file.

    Returns
    -------
    dict

    """
    sizes = {}
    for line in infile:
        if line[0] == "#":
            continue
        sbjct_token, size = line[:-1].split("\t")[:2]
        sizes[sbjct_token] = int(size)

    return sizes


def forceForwardCoordinates(start, end, strand, length):
    """return forward coordinates.

    If strand is negative, the coordinates in a and b
    will be converted. If they are on the positive
    strand, they will be returned as is.

    Arguments
    ---------
    start : int
        Start coordinate
    end : int
        End coordinate
    strand : string
        Strand of interval. The values of "-", "0", "-1" indicate
        a negative strand.
    length : int
        Length of chromosome.
    """
    if strand in ("-", "0", "-1", 0):
        return length - end, length - start
    else:
        return start, end


def CountGeneFeatures(first_position,
                      alignment,
                      genomic_sequence=None,
                      border_stop_codon=0,
                      stop_codons=("TAG", "TAA", "TGA")):
    """calculate number of genomic features in a peptide to genome
    alignment.

    Note that codons can be split, for example::

        S 0 2 5 0 2 I 0 17541 3 0 2 S 1 2 5 0 2 I 0 27979 3 0 2 S 1 2

    Arguments
    ---------
    first_position : int
         Start of alignment on genome.
    alignment : string
         Alignment in :term:`CIGAR` format, for example from exonerate_.
    genomic_sequence : string
         Genomic sequence for alignment
    border_stop_codon : int
         Number of codons that are ignored at the edges of match
         regions.  border_stop_codon should be divisible by three.
    stop_codons : list
         List of stop codons

    Returns
    -------
    nintrons : int
        Number of introns
    nframeshifts : int
        Number of frameshifts in aligment.
    ngaps : int
        Number of gaps in aligment.
    nsplit : int
        Number of codons split by introns in alignment.
    nstopcodons : int
        Number of stop codons in alignment.
    disruptions : list
       List of disruptions in the prediction. Each
       disruption is a tuple of ( "stop|frameshift", position in protein,
       position in cds, position on genomic sequence).

    """

    current_pos_genome = first_position
    current_pos_cds = 0

    nstopcodons = 0
    nintrons = 0
    nframeshifts = 0
    ngaps = 0
    nsplits = 0

    disruptions = []

    # position of nucleotides for split codons
    partial_codon = []

    first_state = True

    for state, l_protein, l_genome in alignment:

        if state in ("G", "M", "P"):

            # reset split codon. Do a sanity check, but ignore first
            # split codon in the case of exon alignments
            if len(partial_codon) % 3 != 0 and not first_state:
                raise ValueError(
                    "split codon was not multiple of three: "
                    "partial_codon=%s, alignment=%s" %
                    (partial_codon, alignment))
                # print alignment
                # pass

            partial_codon = []
            first_state = False

        if state in ("M", "G"):
            # check for stop codons, ignore the first
            # border_stop_codon nucleotides note: this should be a
            # multiple of three
            if genomic_sequence:

                for x in range(border_stop_codon,
                               l_genome - border_stop_codon,
                               3):
                    y = current_pos_cds + x
                    z = current_pos_genome + x
                    codon = genomic_sequence[z:z + 3].upper()

                    if codon in stop_codons:
                        nstopcodons += 1
                        disruptions.append(("stop",
                                            y, y + 3,
                                            z, z + 3))

        if state == "I":
            nintrons += 1

        elif state == "F":
            nframeshifts += 1
            disruptions.append((
                "frameshift",
                current_pos_cds, current_pos_cds + l_genome,
                current_pos_genome, current_pos_genome + l_genome))

        elif state == "G":
            ngaps += 1

        elif state == "P":
            ngaps += 1

        elif state == "S":

            # I used to ignore alignments that start with a split
            # codon as they might have been incomplet. However, the
            # code below seems to work for split codens if not
            # first_state:
            nsplits += 1

            # check for stop-codons in split codons as well:
            for x in range(l_genome):
                partial_codon.append(
                    (current_pos_cds + x, current_pos_genome + x))

            if border_stop_codon < 3 and genomic_sequence and \
               len(partial_codon) % 3 == 0:

                for x in range(0, len(partial_codon), 3):
                    codon = "".join(genomic_sequence[c[1]]
                                    for c in partial_codon[x:x + 3]).upper()
                    if codon in stop_codons:
                        nstopcodons += 1
                        disruptions.append(
                            ("split-stop",
                             partial_codon[x][
                                 0], partial_codon[x + 2][0],
                             partial_codon[x][1], partial_codon[x + 2][1]))
        # advance in cds
        if state in ("M", "G", "F", "S"):
            current_pos_cds += l_genome

        current_pos_genome += l_genome

    return nintrons, nframeshifts, ngaps, nsplits, nstopcodons, disruptions


def Alignment2String(alignment):
    """convert a tuple alignment to an alignment string.
    """
    return " ".join([" ".join(list(map(str, x))) for x in alignment])


def String2Alignment(source):
    """convert an alignment string to a tuple alignment.
    """

    d = source.split(" ")
    ali = []
    if len(d) < 3:
        return ali

    for x in range(0, len(d), 3):
        ali.append((d[x], int(d[x + 1]), int(d[x + 2])))

    return ali


def GetAlignmentLength(alignment):
    """return Alignment length"""

    q, s = 0, 0
    for state, l_query, l_sbjct in alignment:
        q += l_query
        s += l_sbjct

    return q, s


def Alignment2ExonBoundaries(alignment,
                             query_from=0,
                             sbjct_from=0,
                             add_stop_codon=1):
    """extract exon coordinates from a peptide2genome alignment.

    Arguments
    ---------
    aligment : list
        List of tuples of the alignment in CIGAR format.
    query_from : int
        Start position of alignment on peptide sequence.
    sbjct_from : int
        Start position of alignment on nucleotide sequence.
    add_stop_codon : int
        Add final stop codon to exon boundaries.

    Returns
    -------
    exons : list
        A list of exons. Each exon is a tuple of (query_from,
        query_pos, frame, sbjct_from, sbjct_pos, ali)

    """

    exons = []

    # count in nucleotides for query
    query_from *= 3
    query_pos = query_from
    sbjct_pos = sbjct_from
    frame = 0
    ali = []

    p_as_intron = True
    if p_as_intron:
        ali_states = ("S", "M", "G", "F")
    else:
        ali_states = ("S", "M", "G", "F", "P")

    for state, l_query, l_sbjct in alignment:

        # count as nucleotides
        l_query *= 3

        if state in ali_states:
            ali.append((state, l_query, l_sbjct))

        if state == "M":
            query_pos += l_query

        elif state == "G":
            query_pos += l_query

        elif state == "P":
            # treat as intron
            if p_as_intron:
                frame = query_from % 3
                if frame != 0:
                    frame = 3 - frame
                exons.append(
                    (query_from, query_pos, frame, sbjct_from, sbjct_pos, ali))
                ali = []
                query_from = query_pos
                sbjct_from = sbjct_pos + l_sbjct

            query_pos += l_query

        elif state == "S":
            query_pos += l_sbjct

        elif state == "I":
            pass

        elif state == "5":
            # frame is frame for this exon
            frame = query_from % 3
            if frame != 0:
                frame = 3 - frame
            exons.append(
                (query_from, query_pos, frame, sbjct_from, sbjct_pos, ali))
            ali = []
            query_pos += l_query

        elif state == "3":
            query_pos += l_query
            query_from = query_pos
            sbjct_from = sbjct_pos + l_sbjct

        sbjct_pos += l_sbjct

    # add three for the stop codon:
    if add_stop_codon:
        query_pos += 3

    frame = query_from % 3
    if frame != 0:
        frame = 3 - frame

    exons.append((query_from, query_pos, frame, sbjct_from, sbjct_pos, ali))

    return exons


def RemoveFrameShiftsFromAlignment(row_ali, col_ali, gap_char="-"):
    """remove frame shifts in an alignment.

    Frameshifts are gaps are 1, 2, 4, or 5 residues long.

    >>> RemoveFrameShiftsFromAlignment("ABC-EFG", "AB-DEFG")
    ('ABEFG', 'ABEFG')

    Arguments
    ---------
    row_ali : string
        Alignment string of row.
    col_ali : string
        Alignment string of column.
    gap_char : string
        Gap character to identify aligments.

    Returns
    -------
    new_row_ali : string
        New alignment string for row
    new_col_ali : string
        New aligment string for column
    """

    match_string = "[^%s]%s+[^%s]" %\
        (gap_char, gap_char, gap_char)

    positions = []
    for x in re.finditer(match_string, row_ali):
        positions.append((x.start() + 1, x.end() - 1))

    for x in re.finditer(match_string, col_ali):
        positions.append((x.start() + 1, x.end() - 1))

    positions.sort()
    # cut and paste new alignment
    new_row_ali = []
    new_col_ali = []
    a = 0
    for first, last in positions:
        if (last - first) % 3:
            new_row_ali.append(row_ali[a:first])
            new_col_ali.append(col_ali[a:first])
            a = last

    new_row_ali.append(row_ali[a:])
    new_col_ali.append(col_ali[a:])

    return "".join(new_row_ali), "".join(new_col_ali)


def IsStopCodon(codon, stop_codons=("TAG", "TAA", "TGA")):
    """return True if codon is a known stop codon.
    """
    return codon in stop_codons


def MaskStopCodons(sequence, stop_codons=("TAG", "TAA", "TGA")):
    """mask stop codons in a sequence.

    Stop codons are masked with ``NNN``.

    Arguments
    ---------
    sequence : string
        Nucleotide sequence to mask.
    stop_codons : string
        List of known stop codons.

    Returns
    -------
    masked_sequence : string
    """
    codons = []

    for x in range(0, len(sequence), 3):
        codon = sequence[x:x + 3]
        if codon in stop_codons:
            codon = "NNN"
        codons.append(codon)
    return codons.join("")


def Alignment2DNA(alignment, query_from=0, sbjct_from=0):
    """convert a peptide2genome alignment to a nucleotide2nucleotide
    alignment.

    Instead of peptide coordinates, the alignment will be
    in codon coordinates.

    Arguments
    ---------
    aligment : list
        List of tuples of the alignment in CIGAR format.
    query_from : int
        Start position of alignment on peptide sequence.
    sbjct_from : int
        Start position of alignment on nucleotide sequence.

    Returns
    -------
    alignment : object
       The alignment as an alignlib.AlignmentVector object.
    """

    map_query2sbjct = alignlib_lite.py_makeAlignmentVector()

    # count in nucleotides for query
    query_pos = query_from * 3
    sbjct_pos = sbjct_from

    for state, l_query, l_sbjct in alignment:

        # count as nucleotides
        l_query *= 3

        if state in ("A", "B", "C"):

            if state in ("A"):
                l_query = 0
            elif state in ("B"):
                l_query = 1
            elif state in ("C"):
                l_query = 2

        elif state in ("a", "b", "c"):

            if state in ("a"):
                l_query = 0
            elif state in ("b"):
                l_query = 2
            elif state in ("c"):
                l_query = 1

        elif state == "S":
            l_query = l_sbjct

        if l_query > 0 and l_sbjct > 0:
            alignlib_lite.addDiagonal2Alignment(map_query2sbjct,
                                                query_pos, query_pos +
                                                l_query,
                                                sbjct_pos - query_pos)

        query_pos += l_query
        sbjct_pos += l_sbjct

    return map_query2sbjct


Wobble = {
    "A": "GCW",
    "C": "TGY",
    "D": "GAY",
    "E": "GAR",
    "F": "TTY",
    "G": "GGW",
    "H": "CAY",
    "I": "ATW",
    "K": "AAR",
    "L": "WTW",
    "M": "ATG",
    "N": "AAY",
    "P": "CCW",
    "Q": "CAR",
    "R": "WGW",
    "S": "WWW",
    "T": "ACW",
    'U': "TGA",        # selenocysteine
    "V": "GTW",
    "W": "TGG",
    "X": "WWW",        # X can be stop codon or masked sequence.
    "Y": "TAW",
    "*": "NNN",
    "?": "WWW",        # ? can be stop codon or masked sequence.
    "-": "---",
}

GeneticCode = {
    "TTT": "F",
    "TTC": "F",
    "TTA": "L",
    "TTG": "L",
    "TCT": "S",
    "TCC": "S",
    "TCA": "S",
    "TCG": "S",
    "TAT": "Y",
    "TAC": "Y",
    "TAA": "X",  # stop
    "TAG": "X",  # stop
    "TGT": "C",
    "TGC": "C",
    "TGA": "X",  # stop
    "TGG": "W",
    "CTT": "L",
    "CTC": "L",
    "CTA": "L",
    "CTG": "L",
    "CCT": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "CAT": "H",
    "CAC": "H",
    "CAA": "Q",
    "CAG": "Q",
    "CGT": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "ATT": "I",
    "ATC": "I",
    "ATA": "I",
    "ATG": "M",
    "ACT": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "AAT": "N",
    "AAC": "N",
    "AAA": "K",
    "AAG": "K",
    "AGT": "S",
    "AGC": "S",
    "AGA": "R",
    "AGG": "R",
    "GTT": "V",
    "GTC": "V",
    "GTA": "V",
    "GTG": "V",
    "GCT": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "GAT": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "GGT": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
}

GeneticCodeSeleno = GeneticCode.copy()
GeneticCodeSeleno["TGA"] = "U"

GeneticCodeAA = {
    "TTT": "F",
    "TTC": "F",
    "TTA": "L",
    "TTG": "L",
    "TCT": "S",
    "TCC": "S",
    "TCA": "S",
    "TCG": "S",
    "TAT": "Y",
    "TAC": "Y",
    "TGT": "C",
    "TGC": "C",
    "TGG": "W",
    "CTT": "L",
    "CTC": "L",
    "CTA": "L",
    "CTG": "L",
    "CCT": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "CAT": "H",
    "CAC": "H",
    "CAA": "Q",
    "CAG": "Q",
    "CGT": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "ATT": "I",
    "ATC": "I",
    "ATA": "I",
    "ATG": "M",
    "ACT": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "AAT": "N",
    "AAC": "N",
    "AAA": "K",
    "AAG": "K",
    "AGT": "S",
    "AGC": "S",
    "AGA": "R",
    "AGG": "R",
    "GTT": "V",
    "GTC": "V",
    "GTA": "V",
    "GTG": "V",
    "GCT": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "GAT": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "GGT": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
}

StopCodons = ("TAG", "TAA", "TGA")

GeneticCodeDegenerate = {
    "TC": "S",
    "CT": "L",
    "CC": "P",
    "CG": "R",
    "AC": "T",
    "GT": "V",
    "GC": "A",
    "GG": "G",
}


Degeneracy = {
    "TTT": ("F", 1, 1, 2),
    "TTC": ("F", 1, 1, 2),
    "TTA": ("L", 2, 1, 2),
    "TTG": ("L", 2, 1, 2),
    "TCT": ("S", 1, 1, 4),
    "TCC": ("S", 1, 1, 4),
    "TCA": ("S", 1, 1, 4),
    "TCG": ("S", 1, 1, 4),
    "TAT": ("Y", 1, 1, 2),
    "TAC": ("Y", 1, 1, 2),
    "TGT": ("C", 1, 1, 2),
    "TGC": ("C", 1, 1, 2),
    "TGG": ("W", 1, 1, 1),
    "CTT": ("L", 1, 1, 4),
    "CTC": ("L", 1, 1, 4),
    "CTA": ("L", 2, 1, 4),
    "CTG": ("L", 2, 1, 4),
    "CCT": ("P", 1, 1, 4),
    "CCC": ("P", 1, 1, 4),
    "CCA": ("P", 1, 1, 4),
    "CCG": ("P", 1, 1, 4),
    "CAT": ("H", 1, 1, 2),
    "CAC": ("H", 1, 1, 2),
    "CAA": ("Q", 1, 1, 2),
    "CAG": ("Q", 1, 1, 2),
    "CGT": ("R", 1, 1, 4),
    "CGC": ("R", 1, 1, 4),
    "CGA": ("R", 2, 1, 4),
    "CGG": ("R", 2, 1, 4),
    "ATT": ("I", 1, 1, 3),
    "ATC": ("I", 1, 1, 3),
    "ATA": ("I", 1, 1, 3),
    "ATG": ("M", 1, 1, 1),
    "ACT": ("T", 1, 1, 4),
    "ACC": ("T", 1, 1, 4),
    "ACA": ("T", 1, 1, 4),
    "ACG": ("T", 1, 1, 4),
    "AAT": ("N", 1, 1, 2),
    "AAC": ("N", 1, 1, 2),
    "AAA": ("K", 1, 1, 2),
    "AAG": ("K", 1, 1, 2),
    "AGT": ("S", 1, 1, 2),
    "AGC": ("S", 1, 1, 2),
    "AGA": ("R", 2, 1, 2),
    "AGG": ("R", 2, 1, 2),
    "GTT": ("V", 1, 1, 4),
    "GTC": ("V", 1, 1, 4),
    "GTA": ("V", 1, 1, 4),
    "GTG": ("V", 1, 1, 4),
    "GCT": ("A", 1, 1, 4),
    "GCC": ("A", 1, 1, 4),
    "GCA": ("A", 1, 1, 4),
    "GCG": ("A", 1, 1, 4),
    "GAT": ("D", 1, 1, 2),
    "GAC": ("D", 1, 1, 2),
    "GAA": ("E", 1, 1, 2),
    "GAG": ("E", 1, 1, 2),
    "GGT": ("G", 1, 1, 4),
    "GGC": ("G", 1, 1, 4),
    "GGA": ("G", 1, 1, 4),
    "GGG": ("G", 1, 1, 4),
}

GAP_CHAR = "-"

DegeneracyAA = {
    "F": 2,
    "L": 6,
    "S": 4,
    "Y": 2,
    "C": 2,
    "W": 1,
    "P": 4,
    "H": 2,
    "Q": 2,
    "R": 4,
    "I": 3,
    "M": 1,
    "T": 4,
    "N": 2,
    "K": 2,
    "S": 2,
    "R": 2,
    "V": 4,
    "A": 4,
    "D": 2,
    "E": 2,
    "G": 4,
    "X": 0,
    "-": 0,
    ".": 0,
}

AMBIGUOUS_CODES_NA = {
    'A': 'A',
    'C': 'C',
    'G': 'G',
    'T': 'T',
    'N': 'ACGT',
    'X': 'ACGT',
    'U': 'T',
    'R': 'AG',
    'Y': 'CT',
    'M': 'AC',
    'K': 'GT',
    'S': 'CG',
    'W': 'AT',
    'H': 'ACT',
    'B': 'CGT',
    'V': 'ACG',
    'D': 'AGT'}

REVERSE_AMBIGUOUS_NA = {
    'A': 'A', 'C': 'C', 'G': 'G', 'T': 'T',
    'AA': 'A', 'CC': 'C', 'GG': 'G', 'TT': 'T', 'UU': 'U',
    'AG': 'R', 'GA': 'R',
    'CT': 'Y', 'TC': 'Y',
    'AC': 'M', 'CA': 'M',
    'GT': 'K', 'TG': 'K',
    'CG': 'S', 'GC': 'S',
    'AT': 'W', 'TA': 'W',
    'ACT': 'H', 'ATC': 'H', 'CAT': 'H', 'CTA': 'H', 'TAC': 'H', 'TCA': 'H',
    'CGT': 'B', 'CTG': 'B', 'GCT': 'B', 'GTC': 'B', 'TCG': 'B', 'TGC': 'B',
    'ACG': 'V', 'AGC': 'V', 'CAG': 'V', 'CGA': 'V', 'GAC': 'V', 'GCA': 'V',
    'AGT': 'D', 'ATG': 'D', 'GAT': 'D', 'GTA': 'D', 'TAG': 'D', 'TGA': 'D',
}

ENCODE_GENOTYPE = {
    'A': 'A', 'C': 'C', 'G': 'G', 'T': 'T',
    'AA': 'A', 'CC': 'C', 'GG': 'G', 'TT': 'T', 'UU': 'U',
    'AG': 'r', 'GA': 'R',
    'CT': 'y', 'TC': 'Y',
    'AC': 'm', 'CA': 'M',
    'GT': 'k', 'TG': 'K',
    'CG': 's', 'GC': 'S',
    'AT': 'w', 'TA': 'W',
}

DECODE_GENOTYPE = {
    'A': 'AA',
    'C': 'CC',
    'G': 'GG',
    'T': 'TT',
    'r': 'AG', 'R': 'AG',
    'y': 'CT', 'Y': 'CT',
    'm': 'AC', 'M': 'AC',
    'k': 'GT', 'K': 'GT',
    's': 'CG', 'S': 'CG',
    'w': 'AT', 'W': 'AT',
}


def encodeGenotype(code):
    '''encode genotypes like GG, GA into a one-letter code.
    The returned code is lower case if code[0] < code[1], otherwise
    it is uppercase.
    '''
    return ENCODE_GENOTYPE[code.upper()]


def decodeGenotype(code):
    '''decode single letter genotypes like m, M into two letters.
    This is the reverse operation to :meth:`encodeGenotype`.
    '''
    return DECODE_GENOTYPE[code]


def resolveAmbiguousNA(code):
    '''resolve ambiguous nucleic acid letters.
    '''
    return AMBIGUOUS_CODES_NA[code.upper()]


def resolveReverseAmbiguousNA(genotype):
    '''map a genotype to a single letter amino acid amiguous code,
    for example, CT -> Y.
    '''
    return REVERSE_AMBIGUOUS_NA[genotype.upper()]


def GetMapAA2Codons():
    """returns a map of amino acids to codons

    No stop codons.
    ."""
    map_aa2codons = {}

    for codon, aa in list(GeneticCodeAA.items()):
        if aa not in map_aa2codons:
            map_aa2codons[aa] = []
        map_aa2codons[aa].append(codon)
    return map_aa2codons


def GetDegeneracy(codon):
    return Degeneracy[codon.upper()]


def IsStopCodon(codon, stop_codons=("TAG", "TAA", "TGA")):
    return codon in stop_codons


def MapCodon2AA(codon, is_seleno=False, ignore_n=True):
    '''map a codon to an amino acid using the standard translation
    tables

    The mapping returns gaps as gaps and will return an amino acid
    for incomplete codons if there is unambiguous mapping.

    If ``is_seleno`` is set, the codon is translated for a selenoprotein.

    If ``ignore_n`` is set, codons with ``n`` are returned
    as ``?`` in order to distinguish them from stop codons.

    Amino acids are returned as upper-case letters.
    '''

    codon = codon.upper()

    codon = re.sub("[.-]", "", codon)
    if len(codon) == 0:
        return GAP_CHAR

    if is_seleno:
        code = GeneticCodeSeleno
    else:
        code = GeneticCode

    if codon in code:
        return code[codon]
    elif codon == "---":
        return "-"
    elif codon[:2] in GeneticCodeDegenerate:
        # check for four-fold degenerate codons,
        # if they can be mapped (ENSEMBL does it).
        return GeneticCodeDegenerate[codon[:2]]
    elif ignore_n and ("n" in codon or "N" in codon):
        return "?"
    else:
        return "X"


def Protein2Wobble(s):

    c = []
    for x in s:
        c.append(Wobble[x])
    return "".join(c)


def Alignment2PeptideAlignment(alignment,
                               query_from=0,
                               sbjct_from=0,
                               genomic_sequence=None):
    """convert a Peptide2DNA aligment to a Peptide2Peptide alignment.

    How to handle frameshifts?
    """

    map_query2sbjct = alignlib_lite.py_makeAlignmentVector()

    query_pos = query_from
    sbjct_pos = 0
    sbjct_genome_pos = sbjct_from
    sbjct_residues = []
    codon = ""

    for state, l_query, l_sbjct in alignment:

        query_increment = 0
        sbjct_increment = 0

        if state == "M":

            query_increment = l_query
            sbjct_increment = l_sbjct / 3
            if genomic_sequence:
                codon = genomic_sequence[
                    sbjct_genome_pos:sbjct_genome_pos + l_sbjct]

        elif state == "S":
            if l_query:
                sbjct_increment = 1
                query_increment = 1

            if genomic_sequence:
                codon += genomic_sequence[sbjct_genome_pos:
                                          sbjct_genome_pos + l_sbjct]

        elif state == "G":
            query_increment = l_query
            sbjct_increment = l_sbjct / 3
            if genomic_sequence:
                codon += genomic_sequence[sbjct_genome_pos:
                                          sbjct_genome_pos + l_sbjct]

        elif state == "P":
            # only increment query, sbjct does not advance.
            query_increment = l_query

        if query_increment and sbjct_increment:
            alignlib_lite.py_addDiagonal2Alignment(map_query2sbjct,
                                                   query_pos, query_pos +
                                                   query_increment,
                                                   sbjct_pos - query_pos)

        if sbjct_increment and genomic_sequence:
            for x in range(0, len(codon), 3):
                sbjct_residues.append(MapCodon2AA(codon[x:x + 3]))
            codon = ""

        query_pos += query_increment
        sbjct_pos += sbjct_increment

        sbjct_genome_pos += l_sbjct

    return map_query2sbjct, "".join(sbjct_residues)


def translate(sequence,
              is_seleno=False,
              prefer_lowercase=True,
              ignore_n=False,
              ):
    '''convert DNA sequence to a peptide sequence

    If ``is_seleno`` is set, "TGA" codons are treated as
    encoding for selenocysteine.

    If ``ignore_n`` is set, codons with ``n`` are returned
    as ``?`` in order to distinguish them from stop codons.

    '''
    residues = []

    for x in range(0, len(sequence), 3):
        if prefer_lowercase:
            is_lower = False
            for c in sequence[x:x + 3]:
                is_lower = is_lower or c in "acgtnx"
        else:
            is_lower = True
            for c in sequence[x:x + 3]:
                is_lower = is_lower and c in "acgtnx"

        if is_lower:
            residues.append(MapCodon2AA(sequence[x:x + 3].upper(),
                                        is_seleno=is_seleno,
                                        ignore_n=ignore_n).lower())
        else:
            residues.append(MapCodon2AA(sequence[x:x + 3].upper(),
                                        is_seleno=is_seleno,
                                        ignore_n=ignore_n).upper())

    return "".join(residues)


def TranslateDNA2Protein(*args, **kwargs):
    """convert a DNA sequence to a peptide sequence.
    keep case.

    deprecated - use :meth:`translate` instead.
    """
    return translate(*args, **kwargs)


def Alignment2CDNA(alignment,
                   query_from=0,
                   sbjct_from=0,
                   genome=None,
                   remove_frameshifts=0):
    """build cDNA sequence from genomic fragment and
    return alignment of query to it.
    """

    fragments = []
    sbjct_pos = 0
    map_query2sbjct = alignlib_lite.py_makeAlignmentVector()

    # count in nucleotides for query
    query_pos = query_from * 3
    sbjct_pos = sbjct_from
    # position in cDNA
    cdna_pos = 0
    for state, l_query, l_sbjct in alignment:

        # count as nucleotides
        l_query *= 3

        keep = False

        if state == "M":
            keep = True
        elif state == "S":
            l_query = l_sbjct
            keep = True
        elif state == "F" and not remove_frameshifts:
            keep = True
        elif state == "G":
            if l_sbjct > 0:
                keep = True
        elif state == "P":
            keep = False

        if keep:
            if genome:
                fragments.append(genome[sbjct_pos:sbjct_pos + l_sbjct])

            if l_query > 0 and l_sbjct > 0:
                alignlib_lite.py_addDiagonal2Alignment(map_query2sbjct,
                                                       query_pos,
                                                       query_pos + l_query,
                                                       cdna_pos - query_pos)
            cdna_pos += l_sbjct

        query_pos += l_query
        sbjct_pos += l_sbjct

    return map_query2sbjct, fragments.join("")


def GetExon(exons, first_aa):

    first_na = (first_aa - 1) * 3
    combined_na = max(exons[0].mPeptideFrom, first_na)

    # add not covered positions
    g = exons[0].mGenomeFrom + max(0, combined_na - exons[0].mPeptideFrom)

    # add introns
    e = 0
    while exons[e].mPeptideTo < combined_na:
        e += 1
        g += exons[e].mGenomeFrom - exons[e - 1].mGenomeTo

    return g, e, combined_na / 3 + 1


def Exons2Alignment(exons):
    """build an cigar alignment string from a list of exons.
    """

    exons.sort()

    phase = 0
    alignment = []
    last_to = 0
    nexon = 0

    for this_from, this_to in exons:

        nexon += 1

        # print alignment
        # print this_from, this_to

        lintron = this_from - last_to
        lexon = this_to - this_from

        if lintron < 0:
            raise ValueError("error: invalid intron length: %i" % lintron)

        if last_to:

            if (phase != 0):
                missing = 3 - phase

                alignment.append(("S", 0, phase))

            if lintron >= 4:
                # deal with real introns:
                # at least 4 nucleotides for splice signal
                alignment.append(("5", 0, 2))
                alignment.append(("I", 0, lintron - 4))
                alignment.append(("3", 0, 2))

            else:
                # if phase is 0, regard as frameshift
                if phase == 0:
                    alignment.append(("F", 0, lintron))
                else:
                    # deal with frameshift introns
                    # add dummy entries for splice signals
                    alignment.append(("5", 0, 0))
                    alignment.append(("I", 0, lintron))
                    alignment.append(("3", 0, 0))

            if phase != 0:
                # deal with very small exons

                missing = 3 - phase

                # deal with small exons that can not
                # accomodate a full codon
                if missing > lexon:
                    alignment.append(("S", 1, missing))
                else:
                    alignment.append(("S", 1, missing))
                    phase = 0
                    this_from += missing

                # raise ValueError ("expecting a split codon,
                # but small exon %i-%i can not accomodate phase
                # %i in exons %s " % (this_from, this_to, phase, str(exons)))

        l = this_to - this_from

        if l < 0:
            # note: sometimes the last exon just contains
            # a split codon and the stop codon.
            # Thus: do not raise an error if it is the last exon
            if nexon == len(exons):
                break
            else:
                pass
                # raise ValueError ("error: negative length
                # for aligned residues at exon %i-%i in exons
                # %s" % (this_from, this_to, str(exons)))

        phase = l % 3

        ncodons = (l - phase) / 3
        alignment.append(("M", ncodons, ncodons * 3))

        last_to = this_to

    return alignment


def AlignmentProtein2CDNA(src, exons1=None, exons2=None):
    """convert a peptide alignment to a nucleotide
    alignment.

    multiplies coordinates with 3.
    Insert introns.

    Note: alignment starts at 1
    """
    result = src.getNew()

    # enter correct exons (for example, if alignment is not fully covering or
    # gene structure is incomplete)
    gx, exon_id1, first_x = GetExon(exons1, src.getRowFrom())
    gy, exon_id2, first_y = GetExon(exons2, src.getColFrom())

    # print "x=", gx, exon_id1, first_x
    # print "y=", gy, exon_id2, first_y

    last_x = first_x - 1
    last_y = first_y - 1

    exon1_to = exons1[exon_id1].mGenomeTo
    exon2_to = exons2[exon_id2].mGenomeTo

    # only iterate in alignment
    for x in range(first_x,
                   min(src.getRowTo() + 1, exons1[-1].mPeptideTo / 3)):

        y = src.mapRowToCol(x)
        if not y:
            continue
        if y < first_y:
            continue

        gx += (x - last_x - 1) * 3
        gy += (y - last_y - 1) * 3

        # print "# x=", x, "last_x=", last_x, "y=", y, "last_y=", last_y

        for i in range(0, 3):

            # print i, "\t", gx, exons1[exon_id1].mGenomeTo, exon_id2, "\t",
            # gy, exon_id2, exons2[exon_id2].mGenomeTo

            if gx >= exon1_to:
                exon_id1 += 1
                if exon_id1 < len(exons1):
                    gx += exons1[exon_id1].mGenomeFrom - exon1_to
                    exon1_to = exons1[exon_id1].mGenomeTo

            if gy >= exon2_to:
                exon_id2 += 1
                if exon_id2 < len(exons2):
                    gy += exons2[exon_id2].mGenomeFrom - exon2_to
                    exon2_to = exons2[exon_id2].mGenomeTo

            gx += 1
            gy += 1
            result.addPairExplicit(gx, gy, 0)

        last_x = x
        last_y = y

    return result


def GetDegenerateSites(seq1, seq2,
                       degeneracy=4,
                       position=3):
    """returns two new sequenes containing only degenerate sites.

    Only unmutated positions are counted.
    """

    new_seq1 = []
    new_seq2 = []
    for x in range(0, len(seq1), 3):

        c1 = seq1[x:x + 3]
        c2 = seq2[x:x + 3]

        if c1 in GeneticCodeAA and c2 in GeneticCodeAA:

            if GeneticCodeAA[c1] == GeneticCodeAA[c2]:

                if Degeneracy[c1][position] == degeneracy \
                   and Degeneracy[c2][position] == degeneracy:
                    new_seq1.append(c1[position - 1])
                    new_seq2.append(c2[position - 1])

    return "".join(new_seq1), "".join(new_seq2)


class SequencePairInfo:

    """the first characters are ACGT."""

    def __init__(self):
        self.mNIdentical = 0
        self.mNAligned = 0
        self.mNDifferent = 0
        self.mNTransitions = 0
        self.mNTransversions = 0
        self.mNUnaligned1 = 0
        self.mNUnaligned2 = 0
        self.mAlphabet = 0
        self.mMapChar2Pos = {}
        self.mMatrix = []

    def getGCContent(self):
        """return GC content."""
        if not self.mNAligned:
            raise ValueError("no data for calculating GC content")

        cp = self.mMapChar2Pos['C']
        gp = self.mMapChar2Pos['G']
        # two row and two columns, subtract to avoid double counting
        gc = numpy.sum(self.mMatrix[0:4, cp] +
                       self.mMatrix[0:4, gp] +
                       self.mMatrix[cp, 0:4] +
                       self.mMatrix[gp, 0:4]) \
            - self.mMatrix[cp, cp] \
            - self.mMatrix[gp, gp] \
            - self.mMatrix[cp, gp] \
            - self.mMatrix[gp, cp]

        return float(gc) / 2.0 / self.mNAligned

    def __str__(self):
        return "\t".join(map(str, (self.mNIdentical,
                                   self.mNAligned, self.mNDifferent,
                                   self.mNTransitions, self.mNTransversions,
                                   self.mNUnaligned1, self.mNUnaligned2)))

    def getHeader(self):
        return "\t".join(("identical", "aligned", "different",
                          "transitions", "transversions",
                          "unaligned1", "unaligned2"))


class SequencePairInfoCodons(SequencePairInfo):

    """the first characters are ACGT."""

    def __init__(self):
        SequencePairInfo.__init__(self)
        self.mNNonSynonymous = 0
        self.mNSynonymous = 0

    def __str__(self):

        return "\t".join((SequencePairInfo.__str__(self),
                          str(self.mNNonSynonymous),
                          str(self.mNSynonymous)))

    def getHeader(self):
        return "\t".join((SequencePairInfo.getHeader(self), "nnonsyn", "nsyn"))


def AlignedPair2SubstitutionMatrix(seq1, seq2, alphabet):
    """given a pair of sequences, calculate
    a substitution matrix for the given alphabet.
    """
    if len(seq1) != len(seq2):
        raise ValueError("two sequences of unequal length submitted")

    nchars = len(alphabet)
    matrix = numpy.zeros((nchars, nchars), numpy.int)

    map_char2pos = {}
    for x in alphabet:
        map_char2pos[x] = len(map_char2pos)

    for x in range(len(seq1)):
        try:
            matrix[map_char2pos[seq1[x]], map_char2pos[seq2[x]]] += 1
        except KeyError:
            continue

    return matrix


def CalculatePairIndices(seq1, seq2, gap_char="-", with_codons=False):
    """returns number of idential and transitions/transversions substitutions
    in the alignment.

    If with-codons = True, synonymous and nonsynonymous changes will
    be recorded as well. The routine assumes no frame-shifts and will
    count more than one change as non-synonymous.

    """
    alphabet = "ACGT" + gap_char

    map_char2pos = {}
    for x in alphabet:
        map_char2pos[x] = len(map_char2pos)

    # build coordinates for various substitution subsets
    transitions, transversions = [], []
    for x in ("AG", "GA", "CT", "TC"):
        transitions.append((map_char2pos[x[0]], map_char2pos[x[1]]))

    for x in ("AT", "TA", "GT", "TG", "GC", "CG", "AC", "CA"):
        transversions.append((map_char2pos[x[0]], map_char2pos[x[1]]))

    matrix = AlignedPair2SubstitutionMatrix(seq1, seq2, alphabet)
    matrix_acgt = matrix[0:4, 0:4]

    if with_codons:
        result = SequencePairInfoCodons()
    else:
        result = SequencePairInfo()

    result.mMatrix = matrix
    result.mMapChar2Pos = map_char2pos
    result.mNAligned = numpy.sum(numpy.ravel(matrix_acgt))
    result.mNIdentical = numpy.sum(numpy.trace(matrix_acgt))
    result.mNTransitions = numpy.sum([matrix[x] for x in transitions])
    result.mNTransversions = numpy.sum([matrix[x] for x in transversions])
    result.mNDifferent = result.mNAligned - result.mNIdentical
    result.mNUnaligned1 = numpy.sum(numpy.ravel(matrix[0:4, 4]))
    result.mNUnaligned2 = numpy.sum(numpy.ravel(matrix[4, 0:4]))

    if with_codons:
        nsyn, nnon = 0, 0
        pairs = list(zip(seq1, seq2))
        for x in range(len(pairs)):
            a, b = pairs[x]
            if a != b:

                l = (x // 3) * 3
                c1 = MapCodon2AA(seq1[l:l + 3])
                c2 = MapCodon2AA(seq2[l:l + 3])

                if c1 == GAP_CHAR or c2 == GAP_CHAR:
                    continue

                # print x, a, b, l, c1, c2, seq1[l:l+3], seq2[l:l+3], c1 == c2
                if c1 == c2:
                    nsyn += 1
                else:
                    nnon += 1

        result.mNSynonymous = nsyn
        result.mNNonSynonymous = nnon

    return result


def makeSubstitutionMatrix(type="EMBOSS"):
    """make alignator with DNA substitution matrix.

    EMBOSS style matrix:
    identity = 5
    mismatch = -4
    gop = -16
    gep = -4

    ClustalW style matrix:
    match = 1 mismatch = 0
    gop = -10 gep = -0.1
    """

    matrix = []

    if type == "emboss":
        match = 5
        mismatch = -4
        gop = -16.0
        gep = -4.0
    elif type == "blastn":
        match = 1
        mismatch = -3
        gop = -5.0
        gep = -2.0
    elif type == "clustal":
        match = 1
        mismatch = 0
        gop = -10.0
        gep = -0.1
    elif type == "iub":
        match = 1.9
        mismatch = 0
        gop = -10.0
        gep = -0.1
    elif type == "trans":
        match = 1
        mismatch = -5
        gop = -10.0
        gep = -0.1
    elif type == "backtrans":
        # mismatches have to be penalized - there should be none.
        match = 1
        mismatch = -10
        gop = -10.0
        gep = -10.0
    else:
        raise ValueError("unknown MATRIX")

    # build empty matrix
    for m in range(21):
        matrix.append([mismatch] * 21)

    # A <-> A
    matrix[0][0] = match
    # C <-> C
    matrix[1][1] = match
    # G <-> G
    matrix[5][5] = match
    # T <-> T
    matrix[16][16] = match

    if type == "trans":
        # A <-> G
        matrix[0][5] = matrix[5][0] = match
        # C <-> T
        matrix[1][16] = matrix[16][1] = match
    elif type == "backtrans":
        # AGCT <-> W
        for x in (0, 1, 5, 16):
            matrix[x][18] = matrix[18][x] = match
        # AG <-> V
        for x in (0, 5):
            matrix[x][17] = matrix[17][x] = match
        # CT <-> Y
        for x in (1, 16):
            matrix[x][19] = matrix[19][x] = match
        # N <-> N
        for x in (0, 1, 5, 11, 16, 17, 19):
            matrix[x][11] = matrix[11][x] = match

    handle_tmpfile, filename_tmpfile = tempfile.mkstemp()
    for m in matrix:
        os.write(handle_tmpfile, "\t".join(list(map(str, m))) + "\n")
    os.close(handle_tmpfile)

    smatrix = alignlib_lite.py_readSubstitutionMatrixAA(filename_tmpfile)
    os.remove(filename_tmpfile)
    return smatrix, gop, gep


def CalculateRCSUValuesFromCounts(counts, pseudo_counts=0):
    """calculate RCSU values for codons.

    RCSU = relative frequency / uniform frequency
    """


def CalculateCodonFrequenciesFromCounts(counts, pseudo_counts=0):
    """calculate codon frequencies from codon counts per amino acid.
    pseudo_counts are added if desired.
    """
    map_aa2codons = GetMapAA2Codons()
    weights = {}

    for aa, codons in list(map_aa2codons.items()):

        total_counts = reduce(
            lambda x, y: x + y, (counts[x] + pseudo_counts for x in codons))
        if total_counts > 0:
            for x in codons:
                weights[x] = float(counts[x] + pseudo_counts) / total_counts
        else:
            for x in codons:
                weights[x] = 0.0

    for codon in StopCodons:
        weights[codon] = 0.0

    return weights


def CalculateCAIWeightsFromCounts(counts, pseudo_counts=0):
    """calculate CAI weights from codon counts.
    pseudo_counts are added if desired.
    """
    map_aa2codons = GetMapAA2Codons()
    weights = {}

    for aa, codons in list(map_aa2codons.items()):

        max_counts = max(counts[x] for x in codons) + pseudo_counts

        for x in codons:
            weights[x] = float(counts[x] + pseudo_counts) / float(max_counts)

    for codon in StopCodons:
        weights[codon] = 0.0

    return weights


def IsJunk(contig):
    """returns true, if contigs is likely to be junk.

    This is done by name matching. Junk contigs contain either
    one of the following:

    random, unknown, chrU, chU.
    """

    c = contig.lower()
    negative_list = "random", "unknown", "chru", "chu"

    for x in negative_list:
        if re.search(x, c):
            return True

    return False


def CountCodons(sequence):
    """count the codons in a sequence."""
    if len(sequence) % 3 != 0:
        raise "sequence is not multiple of 3: %i" % len(sequence)

    counts = {}
    for c in list(GeneticCodeAA.keys()):
        counts[c] = 0
    for codon in [sequence[x:x + 3] for x in range(0, len(sequence), 3)]:
        try:
            counts[codon] += 1
        except KeyError:
            # skip stop-codons
            pass

    return counts


def GetUniformCodonUsage():
    """get list of frequencies for codons expected for uniform codon usage."""

    frequencies = {}

    aas = {}

    for codon, aa in list(GeneticCodeAA.items()):
        if aa not in aas:
            aas[aa] = 0
        aas[aa] += 1

    for codon, aa in list(GeneticCodeAA.items()):
        frequencies[codon] = 1.0 / aas[aa]

    return frequencies


def GetBiasedCodonUsage(bias=1.0):
    """get list of frequencies for codons according to some bias.

    The first codon for each aa is the most biased, all others are less biased.

    The ratio determines the relative bias between the first and all other
    codons. 0.0 is no bias, 1.0 is complete bias.
    """

    frequencies = {}

    aas = {}

    for codon, aa in list(GeneticCodeAA.items()):
        if aa not in aas:
            aas[aa] = 0
        aas[aa] += 1

    for codon, aa in list(GeneticCodeAA.items()):
        frequencies[codon] = 1.0 / aas[aa]

    return frequencies


def IsNegativeStrand(strand):
    return str(strand) in ("-", "0", "-1")


def IsPositiveStrand(strand):
    return not IsNegativeStrand(strand)


def convertStrand(strand):
    """convert various strand notations into [+-.].
    """
    s = str(strand)
    if s in ("-", "0", "-1"):
        return "-"
    elif s in ("+", "1"):
        return "+"
    else:
        return "."

INTRON_TYPES = (("U2-GT/AG", "GT", "AG"),
                ("U2-nc-GC/AG", "GC", "AG"),
                ("U12-AT/AC", "AT", "AC"))


def GetIntronType(sequence, both_strands=False):
    """return intron type for an intronic sequence.

    If both_strands is True, both strands are checked.
    """

    if both_strands is False:
        for name, prime5, prime3 in INTRON_TYPES:
            if sequence[:len(prime5)].upper() == prime5 and \
                    sequence[-len(prime3):].upper() == prime3:
                return name, prime5, prime3
        else:
            return "unknown", sequence[:5], sequence[-5:]
    else:
        r = complement(sequence)

        for name, prime5, prime3 in INTRON_TYPES:
            if sequence[:len(prime5)].upper() == prime5 and \
                    sequence[-len(prime3):].upper() == prime3:
                return name, prime5, prime3
            elif r[:len(prime5)].upper() == prime5 and \
                    r[-len(prime3):].upper() == prime3:
                return name, prime3, prime5
        else:
            return "unknown", sequence[:5], sequence[-5:]


def printPrettyAlignment(seq1, *args):
    """print a pretty alignment."""

    for other in args:
        assert len(seq1) == len(other)

    seqrow, otherrows = [], []
    nother = len(args)
    for x in range(nother):
        otherrows.extend([[], []])

    for p in range(len(seq1)):
        if p > 0 and p % 3 == 0:
            seqrow.append(" ")
            for x in range(2 * nother):
                otherrows[x] += " "
            if len(seqrow) > 120:
                print("".join(seqrow))
                for x in range(2 * nother):
                    print("".join(otherrows[x]))

                print()
                seqrow, otherrows = [], []
                nother = len(args)
                for x in range(nother):
                    otherrows.extend([[], []])

        l1 = len(seq1[p])
        c1 = seq1[p]
        max_len = max([len(x[p]) for x in args])
        seqrow.append(c1 + " " * (max_len - 11))

        for x in range(nother):

            c2 = args[x][p]
            l2 = len(c2)
            if c1 == c2:
                code = "|"
            elif l1 < l2:
                code = "-"
            elif l2 > l2:
                code = "+"
            else:
                code = ":"
            otherrows[x * 2].append(code + " " * (max_len - len(code)))

            if l2 == 0:
                c2, l2 = "-", 1
            otherrows[x * 2 + 1].append(c2 + " " * (max_len - l2))

    print("".join(seqrow))
    for x in range(2 * nother):
        print("".join(otherrows[x]))


def ReadPeptideSequences(infile, filter=None, as_array=False,
                         regex_identifier=None):
    """read peptide sequence from fasta infile.
    """
    sequences = ParseFasta2Hash(
        infile, filter, regex_identifier=regex_identifier)

    if not as_array:
        for k in list(sequences.keys()):
            sequences[k] = sequences[k][:]
    return sequences


def ParseFasta2Hash(infile, filter=None, regex_identifier=None):
    """read fasta formatted sequences file and build a hash.

    Keys are all characters before the first whitespace in the
    description line.

    Previously, if the key contained a ":", everything before the ":"
    was removed.  This is not true any more.

    Use array for higher space efficiency.

    If regex_identifier is given, this is used to extract the identifier
    from the fasta description line.

    """
    parsed = {}
    key = None
    p = AString.AString()
    if regex_identifier:
        rx = regex_identifier
    else:
        rx = re.compile("^(\S+)")

    for line in infile:
        if line[0] == "#":
            continue
        if not line:
            continue

        if line[0] == ">":
            if key:
                if not filter or key in filter:
                    parsed[key] = p

            key = rx.search(line[1:-1]).groups()[0]
            p = AString.AString()
            continue

        p.extend(AString.AString(re.sub("\s", "", line[:-1])))

    if not filter or key in filter:
        if key:
            parsed[key] = p
    return parsed



'''
fasta2kmercontent.py
=============================================

:Tags: Genomics Sequences FASTA Summary

Purpose
-------

This script takes an input :term:`fasta` file from stdin and computes a
k-nucleotide content for each contig in the file. The output is a
tab-delimited file of kmer counts::

         contig1  contig2  contig3  contig4
    n1
    n2
    n3

where n is the kmer and contig is the fasta entry.

The user specifies the kmer length that is to be searched. Note that the longer
the kmer, the longer the script will take to run.

Note the order of output will not necessarily be the same order as the input.

Usage
-----
Example::

   zcat in.fasta.gz | head::

    >NODE_1_length_120_cov_4.233333
    TCACGAGCACCGCTATTATCAGCAACTTTTAAGCGACTTTCTTGTTGAATCATTTCAATT
    GTCTCCTTTTAGTTTTATTAGATAATAACAGCTTCTTCCACAACTTCTACAAGACGGAAG
    CGTTTTGTAGCTGAAAGTGGGCGAGTTTCCATGATACGAACGATATCGCC

    >NODE_3_length_51_cov_33.000000
    CGAGTTTCCATGATACGAACGATATCGCCTTCTTTAGCAACGTTGTTTTCGTCATGTGCT
    TTATATTTTTTAGAATAGTTGATACGTTTACCATAGACTGG

   zcat in.fasta.gz | python fasta2kmercontent.py
                      --kmer-size 4
                      > tetranucleotide_counts.tsv

   head tetranucleotide_counts.tsv::

     kmer NODE_228_length_74_cov_506.432434 NODE_167_length_57_cov_138.438599
     GTAC 0                                 0
     TGCT 0                                 0
     GTAA 2                                 0
     CGAA 1                                 1
     AAAT 1                                 0
     CGAC 0                                 0

In this example, for each contig in in.fasta.gz the occurrence of each four
nucleotide combination is counted.

Alternative example::

   zcat in.fasta.gz | python fasta2kmercontent.py
                      --kmer-size 4
                      --output-proportion
                      > tetranucleotide_proportions.tsv

In this example, for each contig in in.fasta.gz we return the proportion of
each four base combination out of the total tetranucleotide occurences.
``--output-proportion`` overides the count output.

Options
-------
Two options control the behaviour of fasta2kmercontent.py; ``--kmer-size`` and
``--output-proportion``.

``--kmer-size``::
  The kmer length to count over in the input fasta file

``--output-proportion``::
  The output values are proportions rather than absolute counts


Type::

   python fasta2composition.py --help

for command line help.


Command line options
--------------------

'''

import sys
import re
import CGAT.FastaIterator as FastaIterator
import itertools
import CGAT.Experiment as E


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-k", "--kmer-size", dest="kmer", type="int",
                      help="supply kmer length")

    parser.add_option(
        "-p", "--output-proportion", dest="proportion", action="store_true",
        help="output proportions - overides the default output")

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    # do not allow greater than octonucleotide
    assert options.kmer <= 8, "cannot handle kmer of length %i" % options.kmer

    # how we deal with the nucleotides depends on the kmer length
    nucleotides = []
    for nucleotide in ["A", "C", "T", "G"]:
        nucleotides = nucleotides + \
            [x for x in itertools.repeat(nucleotide, options.kmer)]

    E.info("retrieving %imer sequences" % options.kmer)
    # get all kmer sequences to query
    kmers = set()
    for kmer in itertools.permutations(nucleotides, options.kmer):
        kmers.add(kmer)

    E.info("matching %imers in file" % options.kmer)
    # count the number of kmers in each sequence

    result = {}

    # NB assume that non fasta files are caught by FastaIterator
    total_entries = 0
    for fasta in FastaIterator.iterate(options.stdin):
        total_entries += 1
        result[fasta.title] = {}
        for kmer in kmers:
            counts = [m.start()
                      for m in re.finditer("".join(kmer), fasta.sequence)]
            result[fasta.title][kmer] = len(counts)

    E.info("writing results")
    # write out the results
    headers = sorted(result.keys())
    rows = set()
    for kmer_counts in list(result.values()):
        for kmer, count in kmer_counts.items():
            rows.add("".join(kmer))

    # write header row
    options.stdout.write("kmer\t" + "\t".join(headers) + "\n")

    # output proportions if required - normalises by
    # sequence length
    E.info("computing total counts")
    totals = {}
    for header in headers:
        totals[header] = sum([result[header][tuple(row)] for row in rows])

    for row in sorted(rows):
        if options.proportion:
            options.stdout.write("\t".join(
                [row] + [str(float(result[header][tuple(row)]) / totals[header]) for header in headers]) + "\n")
        else:
            options.stdout.write(
                "\t".join([row] + [str(result[header][tuple(row)]) for header in headers]) + "\n")

    E.info("written kmer counts for %i contigs" % total_entries)
    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
GO.py - compute GO enrichment from gene lists
=============================================

:Tags: Python

Code
----

'''
import sys
import re
import math
import random
import collections

import scipy
import scipy.stats
import scipy.special
import numpy
from CGAT import Stats as Stats
from CGAT import Experiment as E
from CGAT import IOTools as IOTools
from CGAT import Database as Database
from CGAT import CSV as CSV

from rpy2.robjects import r as R

MIN_FLOAT = sys.float_info.min


def lnchoose(n, m):
    nf = scipy.special.gammaln(n + 1)
    mf = scipy.special.gammaln(m + 1)
    nmmnf = scipy.special.gammaln(n - m + 1)
    return nf - (mf + nmmnf)


def hypergeometric_gamma(k, n1, n2, t):
    if t > n1 + n2:
        t = n1 + n2
    if k > n1 or k > t:
        return 0
    elif t > n2 and ((k + n2) < t):
        return 0
    else:
        c1 = lnchoose(n1, k)
        c2 = lnchoose(n2, t - k)
        c3 = lnchoose(n1 + n2, t)

    # print "hyperg:", k, n1, n2, t, math.exp(c1 + c2 - c3)
    return max(math.exp(c1 + c2 - c3), MIN_FLOAT)


def hypergeometric_P(k, n0, n1, t):

    GSL_DBL_EPSILON = 1e-10

    assert t <= (n0 + n1), "t larger than population size"
    assert n0 >= 0, "n0 < 0"
    assert n1 >= 0, "n1 < 0"

    if k >= n0 or k >= t:
        P = 1.0
    elif (k < 0.0):
        P = 0.0
    else:
        P = 0.0
        mode = int(float(t * n0) / float(n0 + n1))
        relerr = 1.0
        if k < mode:
            i = k
            relerr = 1.0
            while(i >= 0 and relerr > GSL_DBL_EPSILON and P < 1.0):
                tmp = hypergeometric_gamma(i, n0, n1, t)
                P += tmp
                relerr = tmp / P
                i -= 1
        else:
            i = mode
            relerr = 1.0
            while(i <= k and relerr > GSL_DBL_EPSILON and P < 1.0):
                tmp = hypergeometric_gamma(i, n0, n1, t)
                P += tmp
                relerr = tmp / P
                i += 1
            i = mode - 1
            relerr = 1.0
            while(i >= 0 and relerr > GSL_DBL_EPSILON and P < 1.0):
                tmp = hypergeometric_gamma(i, n0, n1, t)
                P += tmp
                relerr = tmp / P
                i -= 1
    return P


def hypergeometric_Q(k, n0, n1, t):

    GSL_DBL_EPSILON = 1e-10
    assert t <= (n0 + n1), "t larger than population size"
    assert n0 >= 0, "n0 < 0"
    assert n1 >= 0, "n1 < 0"
    if k >= n0 or k >= t:
        P = 1.0
    elif (k < 0.0):
        P = 0.0
    else:
        P = 0.0
        mode = int(float(t * n0) / float(n0 + n1))
        relerr = 1.0
        if k < mode:
            i = mode
            relerr = 1.0
            while(i <= t and relerr > GSL_DBL_EPSILON and P < 1.0):
                tmp = hypergeometric_gamma(i, n0, n1, t)
                P += tmp
                relerr = tmp / P
                i += 1
            i = mode - 1
            relerr = 1.0
            while(i > k and relerr > GSL_DBL_EPSILON and P < 1.0):
                tmp = hypergeometric_gamma(i, n0, n1, t)
                P += tmp
                relerr = tmp / P
                i -= 1

        else:
            i = k + 1
            relerr = 1.0
            while(i <= t and relerr > GSL_DBL_EPSILON and P < 1.0):
                tmp = hypergeometric_gamma(i, n0, n1, t)
                P += tmp
                relerr = tmp / P
                i += 1
    return P


class Error(Exception):

    """Base class for exceptions in this module."""

    def __str__(self):
        return str(self.message)

    def _get_message(self, message):
        return self._message

    def _set_message(self, message):
        self._message = message
    message = property(_get_message, _set_message)


class InputError(Error):

    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class GOEntry:

    mNameSpaceMap = {
        'molecular_function': 'mol_function',
        'cellular_component': 'cell_location',
        'biological_process': 'biol_process',
    }

    def __init__(self, default_namespace="ontology"):
        self.mNameSpace = default_namespace

    def fromOBO(self, section):
        """read entry form an OBO formatted file."""

        self.mIsA = []

        for line in section:

            data = line[:-1].split(":")
            term = data[0]
            rest = ":".join(data[1:]).strip()
            if term == "name":
                self.mName = rest
            elif term == "id":
                self.mId = rest
            elif term == "namespace":
                self.mNameSpace = self.mNameSpaceMap.get(rest, rest)
            elif term == "def":
                self.mDefinition = rest
            elif term == "exact_synonym":
                self.mSynonym = rest
            elif term == "is_a":
                self.mIsA.append(rest)
            elif term == "comment":
                self.mComment = rest
            elif term == "is_obsolete":
                self.mIsObsolete = True


def readOntology(infile):
    """read ontology in OBO format from infile.

    returns a dictionary of Ontology entries.
    """
    result = {}

    def iterate_blocks(infile):

        lines = []

        for line in infile:
            if line.strip() == "":
                if lines:
                    yield lines
                lines = []
                continue

            lines.append(line)

    default_namespace = "ontology"

    for section in iterate_blocks(infile):

        if section[0].startswith("[Term]"):
            go = GOEntry(default_namespace=default_namespace)
            go.fromOBO(section)
            result[go.mId] = go
        else:
            for line in section:
                data = line[:-1].split(":")
                if data[0] == "default-namespace":
                    default_namespace = data[1].strip()

    return result


class GOSample:

    """store results from sampling.
    """

    def __init__(self, mmin, mmax, mmean, mstddev, mprobovers, mprobunders, counts):

        self.mMin = mmin
        self.mMax = mmax
        self.mMean = mmean
        self.mStddev = mstddev
        self.mProbabilitiesOverRepresentation = mprobovers
        self.mProbabilitiesUnderRepresentation = mprobunders
        self.mCounts = counts


class GOResult:

    mIsOverRepresented = False
    mGOId = None
    mSampleCountsCategory = 0
    mBackgroundCountsCategory = 0
    mSampleCountsTotal = 0
    mBackgroundCountsTotal = 0
    mProbabilityOverRepresentation = 0
    mProbabilityUnderRepresentation = 0
    mPValue = 1.0

    def __init__(self, goid=None):
        self.mGOId = goid

    def UpdateProbabilities(self):
        """calculate probabilities for given counts.

        """
        if self.mBackgroundCountsTotal == 0:
            return

        # various sanity checs
        assert self.mBackgroundCountsCategory >= self.mSampleCountsCategory, \
            "%s: more counts in foreground (%i) than in the background (%i) - make sure the foreground is part of the background." %\
            (self.mGOId, self.mSampleCountsCategory,
             self.mBackgroundCountsCategory)

        assert self.mBackgroundCountsTotal >= self.mBackgroundCountsCategory, \
            "%s: background: more counts in category (%i) than in total (%i)." %\
            (self.mGOId, self.mBackgroundCountsCategory,
             self.mBackgroundCountsTotal)

        assert self.mSampleCountsTotal >= self.mSampleCountsCategory, \
            "%s: forerground: more counts in category (%i) than in total (%i)." %\
            (self.mGOId, self.mSampleCountsCategory, self.mSampleCountsTotal)

        if self.mSampleCountsCategory == 0:
            self.mProbabilityOverRepresentation = 1.0
        else:
            self.mProbabilityOverRepresentation = hypergeometric_Q(self.mSampleCountsCategory - 1,
                                                                   self.mBackgroundCountsCategory,
                                                                   self.mBackgroundCountsTotal -
                                                                   self.mBackgroundCountsCategory,
                                                                   self.mSampleCountsTotal)

        self.mProbabilityUnderRepresentation = hypergeometric_P(self.mSampleCountsCategory,
                                                                self.mBackgroundCountsCategory,
                                                                self.mBackgroundCountsTotal -
                                                                self.mBackgroundCountsCategory,
                                                                self.mSampleCountsTotal)

        self.mPValue = min(
            self.mProbabilityOverRepresentation, self.mProbabilityUnderRepresentation)

        if self.mSampleCountsTotal == 0 or self.mBackgroundCountsCategory == 0:
            self.mRatio = "na"
        else:
            self.mRatio = float(self.mSampleCountsCategory) * self.mBackgroundCountsTotal / \
                self.mSampleCountsTotal / self.mBackgroundCountsCategory

    def getHeaders(self):
        return ["scount", "stotal", "spercent",
                "bcount", "btotal", "bpercent",
                "ratio",
                "pvalue", "pover", "punder"]

    def __str__(self):
        """return string representation."""
        return "%i\t%i\t%s\t%i\t%i\t%s\t%s\t%6.4e\t%6.4e\t%6.4e" % \
            (self.mSampleCountsCategory,
             self.mSampleCountsTotal,
             IOTools.prettyPercent(
                 self.mSampleCountsCategory, self.mSampleCountsTotal),
             self.mBackgroundCountsCategory,
             self.mBackgroundCountsTotal,
             IOTools.prettyPercent(
                 self.mBackgroundCountsCategory, self.mBackgroundCountsTotal),
             IOTools.val2str(self.mRatio),
             self.mPValue,
             self.mProbabilityOverRepresentation,
             self.mProbabilityUnderRepresentation)


class GOResults:

    '''container for go results.'''

    def __init__(self):
        # dictionary of (GOID,GoResult) tuples
        self.mResults = {}
        self.mNumGenes = 0
        self.mBackgroundCountsTotal = 0
        self.mSampleCountsTotal = 0

    def __str__(self):
        """return string representation."""
        lines = []
        lines.append("\t".join(
            map(str, (self.mNumGenes, self.mBackgroundCountsTotal, self.mSampleCountsTotal))))
        for k, v in list(self.mResults.items()):
            lines.append("%s\t%s" % (k, str(v)))
        return "\n".join(lines)


class GOInfo:
    mGOId = None
    mGOType = None
    mDescription = None

    def __init__(self,
                 goid=None,
                 go_type=None,
                 description=None):

        self.mDescription = description
        self.mGOId = goid
        self.mGOType = go_type

    def __str__(self):
        if self.mGOId is None:
            return "\t".join(map(str, ("", "", "")))
        else:
            return "\t".join(map(str, (self.mGOId, self.mGOType, self.mDescription)))

    def getHeaders(self):
        return ["goid", "go_catagory", "go_description"]


class GOMatch(GOInfo):
    mEvidence = None

    def __init__(self,
                 goid=None,
                 go_type=None,
                 description=None,
                 evidence=None):

        GOInfo.__init__(self, goid, go_type, description)
        self.mEvidence = evidence

    def __str__(self):
        return "\t".join(map(str, (self.mGOId, self.mGOType, self.mDescription, self.mEvidence)))


def FilterByGOIds(gene2go, go2info):
    """
    filter gene_id to go_id lookup by a list of go_ids

    returns a new gene2go mapping.

    used to restrict GO terms to GO_slim and remove alternates

    gene2go 				# starting set, map of genes to go terms
    go2info				# alt ids are repeats of superceding ids
    """

    filtered_gene2go = {}

    for gene_id in list(gene2go.keys()):
        new_go = set()
        for go in gene2go[gene_id]:
            if go.mGOId in go2info:
                new_go.add(go)

        if new_go:
            filtered_gene2go[gene_id] = list(new_go)

    return filtered_gene2go


def MapGO2Slims(gene2go, go2slim, ontology=None):
    """filter gene2go lookup by a list of go_ids in go2slim.

    gene2go: map of genes to go terms
    go2slim: map of go categories to goslim go categories   

    If ontology is given, missing descriptions of go entries
    are added from the ontology.

    returns a new gene2go mapping.
    """

    # build map of go identifiers to go info
    map_go2info = {}
    if ontology:
        for go in list(ontology.values()):
            map_go2info[go.mId] = GOInfo(goid=go.mId,
                                         go_type=go.mNameSpace,
                                         description=go.mName)
    else:
        for gene_id, gos in list(gene2go.items()):
            for go in gos:
                map_go2info[go.mGOId] = go

    filtered_gene2go = {}

    for gene_id, gos in list(gene2go.items()):
        new_go = set()
        for go in gos:
            if go.mGOId in go2slim:
                for gg in go2slim[go.mGOId]:
                    if gg in map_go2info:
                        new_go.add(map_go2info[gg])
                    else:
                        raise IndexError(
                            "description for mapped go term not present: %s -> %s" %
                            (go.mGOId, gg))
        if new_go:
            filtered_gene2go[gene_id] = list(new_go)

    return filtered_gene2go


def GetGOSlims(infile):
    """
    returns a map of go identifiers to slim categories

    Input is the output of Chris Mungal's map2slim.pl.
    """

    go2go = {}
    for line in infile:
        if line[:len("part_of")] == "part_of":
            continue

        mapped, parents = line.split("//")
        go, goslims = mapped.split("=>")
        goslims = goslims.split(" ")
        if len(goslims) == 0:
            continue

        go2go[go.strip()] = [x for x in [x.strip() for x in goslims] if len(x)]

    return go2go


def GetGOFrequencies(gene2go, genes):
    """count number of each go category in gene list.

    return a tuple containing:
    * the total number of GO categories found.
    * dictionary of counts per GO category
    * dictionary of genes found with GO categories
    """
    counts = {}
    total = 0
    found_genes = {}

    for gene_id in genes:

        if gene_id not in gene2go:
            continue

        found_genes[gene_id] = 1
        for go in gene2go[gene_id]:
            if go.mGOId not in counts:
                counts[go.mGOId] = 0
            counts[go.mGOId] += 1
            total += 1

    return total, counts, found_genes


def AnalyseGO(gene2go,
              genes,
              genes_background=None,
              do_probabilities=True):
    """analyse go ids.

    goids: list of goids to analyse
    genes: sample set of genes
    genes_background: background set of genes (default: all)
    """
    if genes_background is None:
        genes_background = list(gene2go.keys())

    result = GOResults()

    # get background frequencies
    (background_counts_total, background_counts, background_genes) = \
        GetGOFrequencies(gene2go,
                         genes_background)

    result.mBackgroundCountsTotal = background_counts_total
    result.mBackgroundNumCategories = len(background_counts)
    result.mBackgroundGenes = background_genes

    # get sample frequencies
    (sample_counts_total, sample_counts, sample_genes) = \
        GetGOFrequencies(gene2go,
                         genes)

    result.mNumGenes = len(genes)

    result.mSampleCountsTotal = sample_counts_total
    result.mSampleNumCategories = len(sample_counts)
    result.mSampleGenes = sample_genes

    # test for over or underrepresented categories in the slims
    # report results for all go categories in the background
    # so that also categories completely absent in the foreground (sample)
    # are considered.
    for go_id in list(background_counts.keys()):

        result_go = GOResult(go_id)

        # use gene counts
        result_go.mSampleCountsCategory = sample_counts.get(go_id, 0)
        result_go.mSampleCountsTotal = len(sample_genes)
        result_go.mBackgroundCountsTotal = len(background_genes)
        result_go.mBackgroundCountsCategory = background_counts[go_id]

        E.debug("processing %s: genes in foreground=%i, genes in backgound=%i, sample_counts=%i, background_counts=%i" %
                (go_id,
                 len(sample_genes),
                 len(background_genes),
                 sample_counts.get(go_id, 0),
                 background_counts.get(go_id, 0),
                 )
                )

        if do_probabilities:
            try:
                result_go.UpdateProbabilities()
            except AssertionError as msg:
                print(msg)
                print("# error while calculating probabilities for %s" % go_id)
                print("# genes in sample", sample_genes)
                print("# counts in sample: %i out of %i total" % (
                    result_go.mSampleCountsCategory, result_go.mSampleCountsTotal))
                print("# counts in background %i out of %i total" % (
                    result_go.mBackgroundCountsCategory, result_go.mBackgroundCountsTotal))
                for x in list(sample_genes.keys()):
                    for y in gene2go[x]:
                        print(x, str(y))

                sys.exit(0)

        result.mResults[go_id] = result_go

    return result


def GetGOStatement(go_type, database, species):
    """build statement to get GO assignments for genes from ENSEMBL."""

    if database in ("ensembl_mart_27_1", ):
        statement = """SELECT DISTINCTROW
        gene_stable_id, glook_%s_id, description, olook_evidence_code
        FROM %s.%s_gene_ensembl__go_%s__look
        WHERE glook_%s_id IS NOT NULL
        GROUP BY gene_stable_id, glook_%s_id, description
        ORDER BY gene_stable_id
        """ % (go_type,
               database, species, go_type,
               go_type, go_type)

    elif database in ("ensembl_mart_31", "ensembl_mart_37", "ensembl_mart_41"):
        statement = """SELECT DISTINCTROW
        gene_stable_id, glook_%s_id, description, olook_evidence_code
        FROM %s.%s_go_%s__go_%s__main
        WHERE glook_%s_id IS NOT NULL
        GROUP BY gene_stable_id, glook_%s_id, description
        ORDER BY gene_stable_id
        """ % (go_type,
               database, species, go_type, go_type,
               go_type, go_type)

    elif re.search("core", database):

        if go_type == "biol_process":
            go_type = "biological_process"
        elif go_type == "mol_function":
            go_type = "molecular_function"
        elif go_type == "cell_location":
            go_type = "cellular_component"
        else:
            raise "unknown go_type %s" % go_type

        x = re.search("(\d+)", database)
        if not x:
            raise "can't find version number in database %s" % database

        version = int(x.groups()[0])
        if version <= 54:
            go_database = "ensembl_go_%s" % version
            go_field = "acc"
            statement = """SELECT DISTINCTROW
        g.stable_id, xref.dbprimary_acc, go.name, 'NA'
        FROM gene, transcript, translation, 
        gene_stable_id as g, object_xref as o, xref,
        %(go_database)s.term AS go
        WHERE gene.gene_id = transcript.gene_id
        AND transcript.transcript_id = translation.transcript_id
        AND g.gene_id = gene.gene_id
        AND translation.translation_id = o.ensembl_id
        AND xref.xref_id = o.xref_id
        AND go.%(go_field)s = xref.dbprimary_acc
        AND go.term_type = '%(go_type)s'
        AND xref.external_db_id = 1000
        """ % locals()

        elif version <= 66:
            go_database = "ensembl_ontology_%s" % version
            go_field = "accession"

            statement = """SELECT DISTINCTROW
        g.stable_id, xref.dbprimary_acc, go.name, 'NA'
        FROM gene, transcript, translation, 
        gene_stable_id as g, object_xref as o, xref,
        %(go_database)s.term AS go,
        %(go_database)s.ontology AS ontology
        WHERE gene.gene_id = transcript.gene_id
        AND transcript.transcript_id = translation.transcript_id
        AND g.gene_id = gene.gene_id
        AND translation.translation_id = o.ensembl_id
        AND xref.xref_id = o.xref_id
        AND go.%(go_field)s = xref.dbprimary_acc
        AND go.ontology_id = ontology.ontology_id 
        AND ontology.namespace = '%(go_type)s'
        AND xref.external_db_id = 1000
        """ % locals()

        else:

            go_database = "ensembl_ontology_%s" % version
            go_field = "accession"

            statement = """SELECT DISTINCTROW
        gene.stable_id, xref.dbprimary_acc, go.name, 'NA'
        FROM gene, transcript, translation, 
        object_xref as o, xref,
        %(go_database)s.term AS go,
        %(go_database)s.ontology AS ontology
        WHERE gene.gene_id = transcript.gene_id
        AND transcript.transcript_id = translation.transcript_id
        AND translation.translation_id = o.ensembl_id
        AND xref.xref_id = o.xref_id
        AND go.%(go_field)s = xref.dbprimary_acc
        AND go.ontology_id = ontology.ontology_id 
        AND ontology.namespace = '%(go_type)s'
        AND xref.external_db_id = 1000
        """ % locals()
    else:
        raise "unknown ensmart version %s" % database

    return statement


def ReadGene2GOFromDatabase(dbhandle, go_type, database, species):
    """read go assignments from ensembl database.

    returns a dictionary of lists.
    (one to many mapping of genes to GO categories)
    and a dictionary of go-term to go information

    Note: assumes that external_db_id for GO is 1000
    """

    statement = GetGOStatement(go_type, database, species)
    result = Database.executewait(dbhandle, statement,
                                  retries=0).fetchall()

    gene2go = {}
    go2info = collections.defaultdict(GOInfo)
    for gene_id, goid, description, evidence in result:
        gm = GOMatch(goid, go_type, description, evidence)
        gi = GOInfo(goid, go_type, description)
        if gene_id not in gene2go:
            gene2go[gene_id] = []
        gene2go[gene_id].append(gm)
        go2info[goid] = gi

    return gene2go, go2info


def DumpGOFromDatabase(outfile,
                       dbhandle,
                       options):
    """read go assignments from database.

    and dump them into a flatfile.
    (one to many mapping of genes to GO categories)
    and a dictionary of go-term to go information
    """

    E.info("category\ttotal\tgenes\tcategories")

    all_genes = collections.defaultdict(int)
    all_categories = collections.defaultdict(int)
    all_ntotal = 0

    outfile.write("go_type\tgene_id\tgo_id\tdescription\tevidence\n")

    for go_type in options.ontology:

        genes = collections.defaultdict(int)
        categories = collections.defaultdict(int)
        ntotal = 0
        statement = GetGOStatement(go_type, options.database_name,
                                   options.species)

        results = Database.executewait(
            dbhandle, statement, retries=0).fetchall()

        for result in results:
            outfile.write("\t".join(map(str, (go_type,) + result)) + "\n")
            gene_id, goid, description, evidence = result
            genes[gene_id] += 1
            categories[goid] += 1
            ntotal += 1
            all_genes[gene_id] += 1
            all_categories[goid] += 1
            all_ntotal += 1

        E.info("%s\t%i\t%i\t%i" % (go_type, ntotal,
                                   len(genes),
                                   len(categories)))

    E.info("%s\t%i\t%i\t%i" % ("all",
                               all_ntotal,
                               len(all_genes),
                               len(all_categories)))

    return


def ReadGene2GOFromFile(infile, synonyms={}, obsolete={}):
    """reads GO mappings for all go_types from a
    file.

    If synonyms is given, goids in synynoms will be translated.
    Terms in *obsolete* will be discarded.

    returns two maps: gene2go maps genes to go categories
    and go2info maps go categories to information.
    """

    gene2gos = {}
    go2infos = {}
    c = E.Counter()

    for line in infile:
        if line[0] == "#":
            continue
        try:
            go_type, gene_id, goid, description, evidence = line[
                :-1].split("\t")
        except ValueError as msg:
            raise ValueError("parsing error in line '%s': %s" %
                             (line[:-1], msg))
        if go_type == "go_type":
            continue

        c.input += 1

        if goid in synonyms:
            c.synonyms += 1
            goid = synonyms[goid]

        if goid in obsolete:
            c.obsolete += 1
            continue

        gm = GOMatch(goid, go_type, description, evidence)
        gi = GOInfo(goid, go_type, description)
        if go_type not in gene2gos:
            gene2gos[go_type] = {}
            go2infos[go_type] = {}

        gene2go = gene2gos[go_type]
        go2info = go2infos[go_type]

        if gene_id not in gene2go:
            gene2go[gene_id] = []
        gene2go[gene_id].append(gm)
        go2info[goid] = gi
        c.output += 1

    E.debug("read gene2go assignments: %s" % str(c))

    return gene2gos, go2infos


def CountGO(gene2go):
    """count number of genes and go categories in mapping."""

    cats = collections.defaultdict(int)
    nmaps = 0
    for k, vv in list(gene2go.items()):
        for v in vv:
            nmaps += 1
            cats[v.mGOId] += 1

    return len(gene2go), len(cats), nmaps, cats


def removeCategories(gene2go, categories):
    '''remove all genes that map to *categories*.'''

    for k, vv in list(gene2go.items()):
        gene2go[k] = [v for v in vv if v.mGOId not in categories]


def countGOs(gene2gos):
    """return map of number of genes and go categories in mapping."""
    genes, goids = collections.defaultdict(int), collections.defaultdict(int)

    for cat, gene2go in gene2gos.items():
        for gene_id, vv in gene2go.items():
            genes[gene_id] += 1
            for v in vv:
                goids[v.mGOId] += 1
    return genes, goids


def ReadGeneLists(filename_genes, gene_pattern=None):
    """read gene lists from filename in matrix.

    returns a tuple (list of all genes, dictionary of gene lists) 
    """

    if filename_genes == "-":
        infile = sys.stdin
    else:
        infile = IOTools.openFile(filename_genes, "r")

    headers, table = CSV.readTable(infile.readlines(), as_rows=False)

    if filename_genes != "-":
        infile.close()

    all_genes = table[0]

    # if there is only a single column, add a dummy column
    if len(table) == 1:
        table.append([1] * len(table[0]))
        headers.append("foreground")

    E.info("read %i genes from %s" % (len(all_genes), filename_genes))

    if gene_pattern:
        rx = re.compile(gene_pattern)
        all_genes = [rx.search(x).groups()[0] for x in all_genes]

    gene_lists = collections.OrderedDict()
    for header, col in zip(headers[1:], table[1:]):
        s = list(set([x for x, y in zip(all_genes, col) if y != "0"]))
        gene_lists[header] = set(s)

    return all_genes, gene_lists


def buildGO2Genes(gene2gos, ancestors=None):
    '''invert the dictionary genes2go. 

    If ancestors is given, add missing ancestral information.
    '''

    go2genes = collections.defaultdict(set)
    for gene_id, terms in gene2gos.items():
        for term in terms:
            go2genes[term.mGOId].add(gene_id)
            if ancestors:
                for anc in ancestors[term.mGOId]:
                    go2genes[anc].add(gene_id)
    return go2genes


def GetCode(v):
    """return a code for over/underrepresentation."""

    if v.mRatio > 1.0:
        code = "+"
    elif v.mRatio < 1.0:
        code = "-"
    else:
        code = "?"
    return code


def convertGo2Goslim(options):
    """read gene list with GO assignments and convert to GO slim
    categories."""

    E.info("reading GO assignments from stdin")
    gene2gos, go2infos = ReadGene2GOFromFile(options.stdin)
    input_genes, input_goids = countGOs(gene2gos)

    #############################################################
    # read GO ontology from file
    assert options.filename_ontology, "please supply a GO ontology"
    E.info("reading ontology from %s" % (options.filename_ontology))

    infile = IOTools.openFile(options.filename_ontology)
    ontology = readOntology(infile)
    infile.close()

    go2infos = collections.defaultdict(dict)
    # substitute go2infos
    for go in list(ontology.values()):
        go2infos[go.mNameSpace][go.mId] = GOInfo(go.mId,
                                                 go_type=go.mNameSpace,
                                                 description=go.mName)

    E.info("reading GO assignments from %s" % options.filename_slims)
    go_slims = GetGOSlims(IOTools.openFile(options.filename_slims, "r"))

    if options.loglevel >= 1:
        v = set()
        for x in list(go_slims.values()):
            for xx in x:
                v.add(xx)
        E.info("read go slims from %s: go=%i, slim=%i" %
               (options.filename_slims,
                len(go_slims),
                len(v)))

    output_goids, output_genes = set(), set()
    noutput = 0
    options.stdout.write(
        "\t".join(("go_type", "gene_id", "go_id",
                   "description", "evidence")) + "\n")

    for category, gene2go in sorted(gene2gos.items()):
        gene2go = MapGO2Slims(gene2go, go_slims, ontology)
        for gene_id, values in sorted(gene2go.items()):
            output_genes.add(gene_id)
            for go in sorted(values, key=lambda x: x.mGOId):
                output_goids.add(go.mGOId)
                options.stdout.write("%s\t%s\t%s\t%s\t%s\n" %
                                     (go.mGOType,
                                      gene_id,
                                      go.mGOId,
                                      go.mDescription,
                                      "NA", ))
                noutput += 1

    E.info(
        ("ninput_genes=%i, ninput_goids=%i, noutput_gene=%i, "
         "noutput_goids=%i, noutput=%i") %
        (len(input_genes), len(input_goids),
         len(output_genes), len(output_goids),
         noutput))


def outputResults(outfile,
                  pairs,
                  go2info,
                  options,
                  fdrs=None,
                  samples=None,
                  gene2go=None,
                  foreground=None,
                  gene2name=None):
    '''output GO results to outfile.

    If foreground is given, output a list of gene identifiers in the
    foreground. 

    If gene2name is given, output a columns with gene
    names (instead of identifiers)

    '''

    headers = ["code",
               "scount", "stotal", "spercent",
               "bcount", "btotal", "bpercent",
               "ratio",
               "pvalue", "pover", "punder",
               "goid", "category", "description"]

    if fdrs:
        headers += ["fdr"]

    if gene2go and foreground:
        headers += ['foreground']
        go2genes = buildGO2Genes(gene2go)
        if gene2name:
            headers += ['genes']

    if samples:
        headers += ["min", "max", "zscore", "mpover", "mpunder",
                    "nfdr_expected",
                    "CI95lower", "CI95upper"]

    outfile.write("\t".join(headers) + "\n")

    nselected = 0

    for k, v in sorted(pairs):

        code = GetCode(v)

        n = go2info.get(k, GOInfo())

        outfile.write("%s\t%s\t%s" % (code, str(v), n))

        if options.fdr:
            fdr = fdrs[k][0]
            outfile.write("\t%f" % fdr)

        if options.sample:

            if k in samples:
                s = samples[k]

            # calculate values for z-score
            if s.mStddev > 0:
                zscore = abs(
                    float(v.mSampleCountsCategory) - s.mMean) / s.mStddev
            else:
                zscore = 0.0

            outfile.write("\t%i\t%i\t%f\t%5.2e\t%5.2e\t%6.4f\t%6.4f\t%6.4f" %
                          (s.mMin,
                           s.mMax,
                           zscore,
                           min(s.mProbabilitiesOverRepresentation),
                           min(s.mProbabilitiesUnderRepresentation),
                           scipy.mean(s.mCounts),
                           scipy.stats.scoreatpercentile(s.mCounts, 5),
                           scipy.stats.scoreatpercentile(s.mCounts, 95),
                           ))

        if foreground:
            if k in go2genes:
                g = [x for x in go2genes[k] if x in foreground]
                if gene2name:
                    g = [gene2name.get(x, '?') for x in g]
                g = ";".join(sorted(g))
            else:
                g = ""
            outfile.write("\t%s" % g)

        outfile.write("\n")


def getSamples(gene2go, foreground, background, options, test_ontology,
               go2info):

    sample_size = options.sample
    # List of all minimum probabilities in simulation
    simulation_min_pvalues = []
    E.info("sampling: calculating %i samples: " % (sample_size))

    counts = {}
    prob_overs = {}
    prob_unders = {}

    samples = {}

    options.stdlog.write("# ")
    options.stdlog.flush()

    for x in range(sample_size):

        if options.loglevel >= 1:
            options.stdlog.write(".")
            options.stdlog.flush()

        # get shuffled array of genes from background
        sample_genes = random.sample(background, len(foreground))

        go_results = AnalyseGO(gene2go, sample_genes, background)

        pairs = list(go_results.mResults.items())

        for k, v in pairs:
            if k not in counts:
                counts[k] = []
                prob_overs[k] = []
                prob_unders[k] = []

            counts[k].append(v.mSampleCountsCategory)
            prob_overs[k].append(v.mProbabilityOverRepresentation)
            prob_unders[k].append(v.mProbabilityUnderRepresentation)

            simulation_min_pvalues.append(v.mPValue)

    if options.loglevel >= 1:
        sys.stdout.write("\n")
        sys.stdout.flush()

    E.info("sampling: sorting %i P-Values" % len(simulation_min_pvalues))

    simulation_min_pvalues.sort()
    simulation_min_pvalues = numpy.array(simulation_min_pvalues)

    samples = {}

    if options.output_filename_pattern:
        filename = options.output_filename_pattern % {
            'go': test_ontology, 'section': "samples"}
        E.info("sampling results go to %s" % filename)
        outfile = IOTools.openFile(filename, "w", create_dir=True)
    else:
        outfile = sys.stdout

    outfile.write("\t".join(("goid", "min", "max", "mean", "median", "stddev",
                             "CI95lower", "CI95upper",
                             "pover", "punder", "goid",
                             "category", "description")) + "\n")
    for k in sorted(list(counts.keys())):

        c = counts[k]

        prob_overs[k].sort()
        prob_unders[k].sort()

        s = GOSample(min(c),
                     max(c),
                     scipy.mean(c),
                     numpy.std(c),
                     numpy.array(prob_overs[k]),
                     numpy.array(prob_unders[k]),
                     counts[k])

        samples[k] = s

        outfile.write("%s\t%i\t%i\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%s\n" %
                      (k,
                       min(c),
                       max(c),
                       scipy.mean(c),
                       scipy.median(c),
                       numpy.std(c),
                       scipy.stats.scoreatpercentile(c, 5),
                       scipy.stats.scoreatpercentile(c, 95),
                       min(prob_overs[k]),
                       min(prob_unders[k]),
                       go2info[k]))

    if options.output_filename_pattern:
        outfile.close()

    return samples, simulation_min_pvalues


def computeFDRs(go_results,
                foreground,
                background,
                options,
                test_ontology,
                gene2go,
                go2info):

    pairs = sorted(go_results.mResults.items())

    E.info("calculating the FDRs using method `%s`" % options.qvalue_method)

    samples = None

    observed_min_pvalues = [min(x[1].mProbabilityOverRepresentation,
                                x[1].mProbabilityUnderRepresentation) for x in pairs]

    fdrs = {}

    method = options.qvalue_method

    if options.qvalue_method == "storey":

        # compute fdr via Storey's method
        try:
            fdr_data = Stats.doFDR(observed_min_pvalues)

        except ValueError as msg:
            E.warn("failure in q-value computation: %s" % msg)
            E.warn("reverting to Bonferroni correction")
            method = "bonf"
            fdr_data = Stats.FDRResult()
            l = float(len(observed_min_pvalues))
            fdr_data.mQValues = [min(1.0, x * l) for x in observed_min_pvalues]

        for pair, qvalue in zip(pairs, fdr_data.mQValues):
            fdrs[pair[0]] = (qvalue, 1.0, 1.0)

    elif options.qvalue_method == "empirical":
        assert options.sample > 0, "requiring a sample size of > 0"

        #######################################################################
        # sampling
        # for each GO-category:
        # get maximum and minimum counts in x samples -> calculate minimum/maximum significance
        # get average and stdev counts in x samples -> calculate z-scores for
        # test set
        samples, simulation_min_pvalues = getSamples(gene2go,
                                                     foreground,
                                                     background,
                                                     options,
                                                     test_ontology,
                                                     go2info)

        # compute P-values from sampling
        observed_min_pvalues.sort()
        observed_min_pvalues = numpy.array(observed_min_pvalues)

        sample_size = options.sample

        for k, v in pairs:

            if k in samples:
                s = samples[k]
            else:
                raise KeyError("category %s not in samples" % k)

            # calculate values for z-score
            if s.mStddev > 0:
                zscore = abs(
                    float(v.mSampleCountsCategory) - s.mMean) / s.mStddev
            else:
                zscore = 0.0

            #############################################################
            # FDR:
            # For each p-Value p at node n:
            #   a = average number of nodes in each simulation run with P-Value < p
            #           this can be obtained from the array of all p-values and all nodes
            #           simply divided by the number of samples.
            #      aka: expfpos=experimental false positive rate
            #   b = number of nodes in observed data, that have a P-Value of less than p.
            #      aka: pos=positives in observed data
            #   fdr = a/b
            pvalue = v.mPValue

            # calculate values for FDR:
            # nfdr = number of entries with P-Value better than node.
            a = 0
            while a < len(simulation_min_pvalues) and \
                    simulation_min_pvalues[a] < pvalue:
                a += 1
            a = float(a) / float(sample_size)
            b = 0
            while b < len(observed_min_pvalues) and \
                    observed_min_pvalues[b] < pvalue:
                b += 1

            if b > 0:
                fdr = min(1.0, float(a) / float(b))
            else:
                fdr = 1.0

            fdrs[k] = (fdr, a, b)
    else:
        qvalues = R['p.adjust'](
            observed_min_pvalues, method=options.qvalue_method)
        fdr_data = Stats.FDRResult()
        fdr_data.mQValues = list(qvalues)
        for pair, qvalue in zip(pairs, fdr_data.mQValues):
            fdrs[pair[0]] = (qvalue, 1.0, 1.0)

    return fdrs, samples, method


def getFileName(options, **kwargs):
    '''return a filename

    Placeholders in filename are string-substituted with the 
    dictionary in kwargs.
    '''
    if options.output_filename_pattern:
        filename = options.output_filename_pattern % kwargs
        E.info("output for section '%s' go to %s" %
               (kwargs.get("section", "unknown"), filename))
        outfile = IOTools.openFile(filename, "w", create_dir=True)
    else:
        outfile = options.stdout

    return outfile


def buildMatrix(results, valuef, dtype=numpy.float, default=0):
    '''build a matrix from a field in *results*

    The value stored in the matrix is accessed via *valuef*.
    '''

    row_headers = [set([x[0] for x in y]) for y in results]
    row_headers = sorted(list(row_headers[0].union(*row_headers[1:])))
    map_row = dict(list(zip(row_headers, list(range(len(row_headers))))))
    matrix = numpy.zeros((len(row_headers), len(results)), dtype=dtype)
    if default != 0:
        matrix[:] = default

    for col, pairs in enumerate(results):
        for row, v in pairs:
            try:
                matrix[map_row[row]][col] = valuef(v)
            except ValueError:
                # ignore errors for log(0)
                pass

    return matrix, row_headers


def selectSignificantResults(pairs, fdrs, options):
    '''select a set of significant results.
    '''

    filtered_pairs = []
    for k, v in pairs:

        is_ok = False

        pvalue = v.mPValue

        if options.fdr:
            (fdr, expfpos, pos) = fdrs[k]
            if fdr < options.threshold:
                is_ok = True
        else:
            if pvalue < options.threshold:
                is_ok = True

        if is_ok:
            filtered_pairs.append((k, v))

    return filtered_pairs


def outputMultipleGeneListResults(results,
                                  all_genelists_with_results,
                                  test_ontology,
                                  go2info,
                                  options,
                                  section):
    '''select a set of significant results.
    '''

    col_headers = all_genelists_with_results

    if len(results) == 0:
        E.warn('no significant results - no matrices output')
        return

    assert len(col_headers) == len(results)

    def _output(section, subsection, valuef, dtype):

        # fold change matrix
        matrix, row_headers = buildMatrix(results,
                                          valuef=valuef,
                                          dtype=dtype)

        outfile = getFileName(options,
                              go=test_ontology,
                              section=section,
                              set='%s_all' % subsection)

        IOTools.writeMatrix(
            outfile, matrix, row_headers, col_headers, row_header="category")

        outfile = getFileName(options,
                              go=test_ontology,
                              section=section,
                              set='%s_alldesc' % subsection)

        IOTools.writeMatrix(outfile, matrix,
                            ["%s:%s" % (x, go2info[x].mDescription)
                             for x in row_headers],
                            col_headers, row_header="category")

    _output('l2fold', section,
            valuef=lambda x: math.log(x.mRatio + 0.00000001, 2),
            dtype=numpy.float)

    _output('l10pvalue', section,
            valuef=lambda x: int(-10 * math.log(x.mPValue, 10)),
            dtype=numpy.int)

    _output('l10qvalue', section,
            valuef=lambda x: int(-10 * math.log(x.mQValue, 10)),
            dtype=numpy.int)


def pairwiseGOEnrichment(results_per_genelist, labels, test_ontology, go2info,
                         options):
    '''compute pairwise enrichment between sets.

    The purpose of this method is to find if there are categories that are differently enriched
    in a pair of gene lists.

    The appropriate test here is the Chi-Squared test. 

    The assumption is that the background set is the same in all gene lists.

    The workflow is thus::

       for each combination of two gene lists:
           for each GO category:
               get counts in foreground, total counts of foreground
               compute chi-square enrichment output
               save P-value
           apply fdr - output significant differences.
    '''

    dicts = [dict(x) for x in results_per_genelist]

    PairResult = collections.namedtuple("PairResult",
                                        "goid set1 set2 counts1 total1 pvalue1 qvalue1 counts2 total2 pvalue2 qvalue2 pvalue qvalue description")

    outfile = getFileName(options,
                          go=test_ontology,
                          section='summary',
                          set="pairs")

    outfile.write(
        "set1\tset2\ttotal1\ttotal2\tshared\tskipped\ttested\tsignificant\tinsignificant\n")

    results = []

    total = len(dicts) * (len(dicts) - 1) / 2

    iteration = 0

    min_observed_counts = options.pairs_min_observed_counts

    for x, genelist1 in enumerate(sorted(dicts)):

        x_go_categories = set(genelist1.keys())
        for y, genelist2 in enumerate(sorted(dicts[:x])):

            iteration += 1
            if iteration % 10 == 0:
                E.info("iteration: %i/%i (%5.2f%%)" %
                       (iteration, total, 100.0 * iteration / total))

            y_go_categories = set(genelist2.keys())

            shared = x_go_categories.intersection(y_go_categories)

            c = E.Counter()

            for category in shared:
                c.shared += 1
                xx = genelist1[category]
                yy = genelist2[category]

                # discard all tests with few observations in the observed
                # counts
                if xx.mSampleCountsCategory < min_observed_counts and yy.mSampleCountsCategory < min_observed_counts:
                    c.skipped += 1
                    continue

                observed = (xx.mSampleCountsCategory, yy.mSampleCountsCategory)

                aa, bb, cc, dd = \
                    (xx.mSampleCountsCategory,
                     yy.mSampleCountsCategory,
                     xx.mSampleCountsTotal - xx.mSampleCountsCategory,
                     yy.mSampleCountsTotal - yy.mSampleCountsCategory)

                if cc == dd == 0:
                    c.skipped += 1
                    continue

                c.tested += 1

                fisher, pvalue = scipy.stats.fisher_exact(numpy.array(
                    ((aa, bb),
                     (cc, dd))))

                if pvalue < 0.05:
                    c.significant_pvalue += 1
                else:
                    c.insignificant_pvalue += 1

                results.append(PairResult._make((category,
                                                 labels[x],
                                                 labels[y],
                                                 xx.mSampleCountsCategory,
                                                 xx.mSampleCountsTotal,
                                                 xx.mPValue,
                                                 xx.mQValue,
                                                 yy.mSampleCountsCategory,
                                                 yy.mSampleCountsTotal,
                                                 yy.mPValue,
                                                 yy.mQValue,
                                                 pvalue,
                                                 1.0,
                                                 go2info[category].mDescription)))

            outfile.write("\t".join(map(str,
                                        (labels[x], labels[y],
                                         len(x_go_categories),
                                         len(y_go_categories),
                                         c.shared,
                                         c.skipped,
                                         c.tested,
                                         c.significant_pvalue,
                                         c.insignicant_pvalue))) + "\n")
    if options.output_filename_pattern:
        outfile.close()

    if options.fdr:
        pvalues = [x.pvalue for x in results]

        if options.qvalue_method == "storey":

            # compute fdr via Storey's method
            try:
                fdr_data = Stats.doFDR(pvalues)

            except ValueError as msg:
                E.warn("failure in q-value computation: %s" % msg)
                E.warn("reverting to Bonferroni correction")
                method = "bonf"
                fdr_data = Stats.FDRResult()
                l = float(len(pvalues))
                fdr_data.mQValues = [min(1.0, x * l) for x in pvalues]

            qvalues = fdr_data.mQValues
        else:
            qvalues = R['p.adjust'](pvalues, method=options.qvalue_method)

        # update qvalues
        results = [x._replace(qvalue=y) for x, y in zip(results, qvalues)]

    outfile = getFileName(options,
                          go=test_ontology,
                          section='pairs',
                          set="pairs")

    outfile.write("\t".join(PairResult._fields) + "\n")
    for result in results:
        outfile.write("\t".join(map(str, result)) + "\n")

    if options.output_filename_pattern:
        outfile.close()



'''Style.py
============

utility functions for checking coding style in
CGAT scripts.

'''

import subprocess
import collections
import re

DATA = collections.namedtuple('DATA', 'count code description')


def runPep8(expr):
    '''run pep8 on files matching the glob expression
    *expr*
    '''

    args = ['/ifs/devel/andreas/python/bin/pep8',
            '--statistics',
            '--quiet',
            expr]

    args = ' '.join(args)
    # pep8 returns unix error if there are errors
    try:
        output = subprocess.check_output(args, shell=True)
    except subprocess.CalledProcessError as msg:
        output = msg.output

    # count python files
    nchecked = len([x for x in output.split('\n') if x.endswith('.py')])

    # parse data
    # Output is '120 W123 Description'
    lines = [x for x in output.split('\n') if not x.endswith('.py')]
    data = [DATA(*re.split(' +', x, 2)) for x in lines if x]

    return nchecked, data



import sys
from pysam.libchtslib cimport *
from pysam.libcalignmentfile cimport *
from pysam.libcfaidx cimport *
from pysam.libcutils cimport force_str, force_bytes
from libc.string cimport strchr
from libc.stdint cimport int8_t
from libc.stdio cimport puts, printf
import collections, array, struct, sys
import CGAT.Experiment as E

import hashlib
import base64

FLAGS = {
    1: 'paired',
    2: 'proper_pair',
    4: 'unmapped',
    8: 'mate_unmapped',
    16: 'reverse',
    32: 'mate_reverse',
    64: 'read1',
    128: 'read2',
    256: 'secondary',
    512: 'qc_fail',
    1024: 'duplicate'}

cdef struct CountsType:
    int8_t is_mapped
    int8_t is_unmapped
    int8_t mate_is_unmapped
    int8_t is_paired
    int8_t mapped_is_read1
    int8_t mapped_is_read2
    int8_t unmapped_is_read1
    int8_t unmapped_is_read2
    int8_t is_proper_pair
    int8_t is_qcfail
    int8_t is_duplicate

def count(AlignmentFile samfile,
          remove_rna,
          rna,
          filename_fastq=None,
          outfile_details=None):
    '''
    '''
    cdef bint _remove_rna = remove_rna

    cdef AlignedSegment read

    # counters
    cdef int ninput = 0
    cdef int nduplicates = 0
    # number of reads overlapping RNA (if rna != None)
    cdef int nrna = 0
    # number of reads not overlap RNA (if rna != None)
    cdef int n_norna = 0
    # number of reads present after filtering
    cdef int nfiltered = 0

    cdef int max_hi = 0
    # count nh, nm tags
    nh_filtered = collections.defaultdict(int)
    nm_filtered = collections.defaultdict(int)
    nh_all = collections.defaultdict(int)
    nm_all = collections.defaultdict(int)
    mapq_filtered = collections.defaultdict(int)
    mapq_all = collections.defaultdict(int)

    cdef int * flags_counts = <int*>calloc(len(FLAGS), sizeof(int))

    # helper variables
    cdef int last_tid = -1
    cdef int last_pos = 0

    duplicates = collections.defaultdict(int)
    counts = collections.defaultdict(int)
    cdef int count = 0

    cdef int tid = -1
    cdef int flag

    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int32_t hi
    cdef int x
    cdef int lflags = len(FLAGS)
    cdef int f

    # detailed counting
    cdef PersistentFastqProxy fq
    cdef int64_t index, fastq_nreads
    cdef CountsType * fastq_counts
    cdef CountsType * fastq_count
    cdef char * read_name
    cdef char * position
    cdef int count_fastq = filename_fastq != None
    cdef int fastq_notfound = 0
    cdef int chop = 0
    
    cdef bint is_py3 = sys.version_info.major >= 3

    if count_fastq:
        E.info("reading fastq file")
        # Using a python dictionary here is due
        # for a large amount of memory usage.
        # Alternatives to dictionary
        # 1. POSIX hash tables (hsearch,...) or trees: very slow
        # 2. custom hash implementation: worth the effort?
        # 3. Sorted list and binary search: too slow for many lookups
        reads = {}
        fastqfile = FastxFile(filename_fastq)
        fastq_nreads = 0
        for fq in fastqfile:
            if fastq_nreads == 0:
                # chop off /1 or /2 as mappers usually remove these
                # suffices. Test only the first.
                name = fq.name
                if name.endswith("/1") or name.endswith("/2"):
                    chop = -2
                
            if chop != 0:
                s = fq.name[:chop]
            else:
                s = fq.name

            name = hashlib.md5(force_bytes(s)).digest()

            reads[name] = fastq_nreads

            fastq_nreads += 1
        
        E.info("read names of %i reads or read pairs" % fastq_nreads)

        E.info("allocating %i bytes" % (fastq_nreads * sizeof(CountsType)))

        fastq_counts = <CountsType *>calloc(fastq_nreads, sizeof(CountsType))
        if fastq_counts == NULL:
            raise ValueError("could not allocate memory for %i bytes" %
                             (fastq_nreads * sizeof(CountsType)))
 
    for read in samfile:

        if count_fastq:
            read_name = pysam_bam_get_qname(read._delegate)
            if read_name == NULL:
                raise ValueError("file does not contain read names, can't count using fastq")

            # terminate string at first space to
            # truncate read names containing more than
            # just the id
            position = strchr(read_name, ' ')
            if position != NULL: 
                position[0] = '\0'

            md5 = hashlib.md5(read_name).digest()
            try:
                fastq_count = &fastq_counts[reads[md5]]
            except KeyError:
                fastq_notfound += 1
                continue

            # book-keeping counts
            if read.is_qcfail: fastq_count.is_qcfail += 1
            if read.is_duplicate: fastq_count.is_duplicate += 1
            if read.is_paired: fastq_count.is_paired += 1
                
            if read.is_unmapped: 
                fastq_count.is_unmapped += 1
                if read.is_read1: fastq_count.unmapped_is_read1 += 1
                if read.is_read2: fastq_count.unmapped_is_read2 += 1
            else:
                # only count is_read1 and is_read2 for mapped
                fastq_count.is_mapped += 1
                if read.mate_is_unmapped: fastq_count.mate_is_unmapped += 1
                if read.is_read1: fastq_count.mapped_is_read1 += 1
                if read.is_read2: fastq_count.mapped_is_read2 += 1
                if read.is_proper_pair: fastq_count.is_proper_pair += 1
 
        flag = read._delegate.core.flag

        # is paired and read2
        #if flag & 1 and flag & 128:
            # ignore second reads to avoid double counting pairs
            # this needs to be done properly by counting in
            # pairs
        #    continue

        ninput += 1

        f = 1
        for x from 0 <= x < lflags:
            if flag & f: flags_counts[x] += 1
            f = f << 1

        # get maximum NI field
        v = bam_aux_get(read._delegate, 'HI')
        if v != NULL:
            hi = <int32_t>bam_aux2i(v)
            if hi > max_hi: max_hi = hi

        v = bam_aux_get(read._delegate, 'NH')
        if v != NULL:
            nh = <int32_t>bam_aux2i(v)
            nh_all[nh] += 1
        else:
            nh = -1

        v = bam_aux_get(read._delegate, 'NM')
        if v != NULL:
            nm = <int32_t>bam_aux2i(v)
            nm_all[nm] += 1
        else:
            nm = -1

        mapq_all[read.mapq] += 1

        # skip unmapped reads
        if read._delegate.core.flag & 4: continue

        if read.tid != last_tid:
            contig = samfile.getrname(read.rname)

        # note: does not take into account gaps within reads
        # or partial overlap.
        if rna:
            if rna.contains(contig, read.pos, read.pos + read.alen):
                nrna += 1
                if _remove_rna:
                    continue
            else:
                n_norna += 1

        nfiltered += 1

        if nh >= 0: nh_filtered[nh] += 1
        if nm >= 0: nm_filtered[nm] += 1
        mapq_filtered[read.mapq] += 1

        # duplicate analysis - simply count per start position
        # ignoring sequence and strand
        if read.tid == last_tid and read.pos == last_pos:
            count += 1
            nduplicates += 1
            continue

        if count > 1:
            counts[count] += 1

        count = 1
        last_tid, last_pos = read.tid, read.pos

    E.info( "finished computing counts" )

    if fastq_notfound:
        E.warn( "could not match %i records in bam-file to fastq file" % fastq_notfound)

    counter = E.Counter()

    counter.alignments_input = ninput
    counter.alignments_filtered = nfiltered
    counter.alignments_rna = nrna
    counter.alignments_no_rna = n_norna
    counter.alignments_duplicates = nduplicates

    # convert flags to labels
    t = {}
    f = 1
    for x from 0 <= x < lflags:
        t[FLAGS[f]] = flags_counts[x]
        f = f << 1

    # count based on fastq data
    cdef int total_paired = 0
    cdef int total_unpaired = 0
    cdef int total_pair_is_mapped = 0
    cdef int total_pair_is_unmapped = 0
    cdef int total_pair_is_proper_uniq = 0
    cdef int total_pair_is_proper_mmap = 0
    cdef int total_pair_is_proper_duplicate = 0
    cdef int total_pair_not_proper_uniq = 0
    cdef int total_pair_is_incomplete_uniq = 0
    cdef int total_pair_is_incomplete_mmap = 0
    cdef int total_pair_is_other = 0
    # read based counting for unpaired reads
    cdef int total_reads = 0
    cdef int total_read_is_mapped_uniquely = 0
    cdef int total_read_is_mapped_multiply = 0
    cdef int total_read_is_unmapped = 0
    cdef int total_read_is_missing = 0
    # read based counting for read1
    cdef int total_read1 = 0
    cdef int total_read1_is_mapped_uniquely = 0
    cdef int total_read1_is_mapped_multiply = 0
    cdef int total_read1_is_unmapped = 0
    cdef int total_read1_is_missing = 0
    # read based counting for read2
    cdef int total_read2 = 0
    cdef int total_read2_is_mapped_uniquely = 0
    cdef int total_read2_is_mapped_multiply = 0
    cdef int total_read2_is_unmapped = 0
    cdef int total_read2_is_missing = 0
    
    if count_fastq:

        E.info("fastq counting: aggregating counts")
        for qname, index in reads.items():
            fastq_count = &fastq_counts[index]

            # paired read data
            if fastq_count.is_paired: 
                total_paired += 1

                if fastq_count.is_unmapped == fastq_count.is_paired:
                    # an unmapped read pair
                    total_pair_is_unmapped += 1
                    continue

                total_pair_is_mapped += 1

                if fastq_count.is_proper_pair == 2: 
                    # a unique proper pair
                    total_pair_is_proper_uniq +=1
                    # a duplicate unique proper pair
                    if fastq_count.is_duplicate == 2:
                        total_pair_is_proper_duplicate +=1
                elif fastq_count.is_proper_pair > 2:
                    # proper pairs that map to multiple locations
                    total_pair_is_proper_mmap +=1
                elif fastq_count.mapped_is_read1 == 1 \
                        and fastq_count.mapped_is_read2 == 1 \
                        and fastq_count.is_proper_pair == 0:
                    # pair which map each read once, but not is not proper
                    total_pair_not_proper_uniq += 1
                elif (fastq_count.mapped_is_read1 == 1 and
                      fastq_count.mapped_is_read2 == 0) \
                    or (fastq_count.mapped_is_read1 == 0 and
                        fastq_count.mapped_is_read2 == 1):
                    # an incomplete pair - one read of a pair matches uniquely 
                    # but not the other
                    total_pair_is_incomplete_uniq += 1
                elif (fastq_count.mapped_is_read1 == 1 and
                      fastq_count.mapped_is_read2 > 1) \
                    or (fastq_count.mapped_is_read1 > 1 and
                        fastq_count.mapped_is_read2 == 1):
                    # an incomplete pair - one read of a pair matches uniquely
                    # but the other matches multiple times
                    total_pair_is_incomplete_mmap += 1
                else:
                    total_pair_is_other += 1

                # paired read counting
                if fastq_count.mapped_is_read1: 
                    if fastq_count.mapped_is_read1 == 1:
                        total_read1_is_mapped_uniquely += 1
                    elif fastq_count.mapped_is_read1 > 1:
                        total_read1_is_mapped_multiply += 1
                elif fastq_count.unmapped_is_read1 > 0:
                    total_read1_is_unmapped += 1
                else:
                    total_read1_is_missing += 1

                if fastq_count.mapped_is_read2: 
                    if fastq_count.mapped_is_read2 == 1:
                        total_read2_is_mapped_uniquely += 1
                    elif fastq_count.mapped_is_read2 > 1:
                        total_read2_is_mapped_multiply += 1
                elif fastq_count.unmapped_is_read2 > 0:
                    total_read2_is_unmapped += 1
                else:
                    total_read2_is_missing += 1
                        
            else:
                # reads without data, could be unpaired or missing
                total_unpaired += 1

                # counting for unpaired data
                if fastq_count.is_unmapped > 0:
                    total_read_is_unmapped += 1
                else:
                    if fastq_count.is_mapped == 1:
                        total_read_is_mapped_uniquely += 1
                    elif fastq_count.is_mapped > 1:
                        total_read_is_mapped_multiply += 1
                    else:
                        total_read_is_missing += 1
                    
        ###################################################
        # Pair statistics
        counter.total_pairs = total_paired + total_unpaired
        counter.total_pair_is_mapped = total_pair_is_mapped
        counter.total_pair_is_unmapped = total_pair_is_unmapped
        counter.total_pair_is_proper_uniq = total_pair_is_proper_uniq
        counter.total_pair_is_incomplete_uniq = total_pair_is_incomplete_uniq
        counter.total_pair_is_incomplete_mmap = total_pair_is_incomplete_mmap
        counter.total_pair_is_proper_duplicate = total_pair_is_proper_duplicate
        counter.total_pair_is_proper_mmap = total_pair_is_proper_mmap
        counter.total_pair_not_proper_uniq = total_pair_not_proper_uniq
        counter.total_pair_is_other = total_pair_is_other

        ###################################################
        # Read Stats for SE and PE data
        # for SE data
        if total_paired == 0:
            E.info("reads: single end counting")
            counter.total_read = fastq_nreads
            counter.total_read_is_mapped_uniq = total_read_is_mapped_uniquely
            counter.total_read_is_mmap = total_read_is_mapped_multiply
            counter.total_read_is_mapped = total_read_is_mapped_multiply +\
                                           total_read_is_mapped_uniquely
            counter.total_read_is_unmapped = total_read_is_unmapped
            counter.total_read_is_missing = total_read_is_missing
        else:
            # for PE data
            E.info("reads: paired end counting: paired = %i" % total_paired)
            counter.total_read = fastq_nreads * 2
            counter.total_read_is_mapped_uniq = total_read1_is_mapped_uniquely +\
                                                total_read2_is_mapped_uniquely
            counter.total_read_is_mmap = total_read1_is_mapped_multiply +\
                                         total_read2_is_mapped_multiply
            counter.total_read_is_mapped = counter.total_read_is_mapped_uniq +\
                                           counter.total_read_is_mmap
            counter.total_read_is_unmapped = total_read1_is_unmapped +\
                                             total_read2_is_unmapped
            counter.total_read_is_missing = total_read1_is_missing +\
                                            total_read2_is_missing

        ###################################################
        ## stats for 1st/2nd read separately
        counter.total_read1 = fastq_nreads
        counter.total_read1_is_mapped_uniq = total_read1_is_mapped_uniquely
        counter.total_read1_is_mmap = total_read1_is_mapped_multiply
        counter.total_read1_is_mapped = counter.total_read1_is_mapped_uniq + counter.total_read1_is_mmap
        counter.total_read1_is_unmapped = total_read1_is_unmapped
        counter.total_read1_is_missing = total_read1_is_missing
        counter.total_read2 = fastq_nreads
        counter.total_read2_is_mapped_uniq = total_read2_is_mapped_uniquely
        counter.total_read2_is_mmap = total_read2_is_mapped_multiply
        counter.total_read2_is_mapped = counter.total_read2_is_mapped_uniq + counter.total_read2_is_mmap
        counter.total_read2_is_unmapped = total_read2_is_unmapped
        counter.total_read2_is_missing = total_read2_is_missing

        if outfile_details:
            if outfile_details != sys.stdout:
                # later: get access FILE * object
                outfile_details.write(
                    "read_md5\tis_unmapped\tmate_is_unmapped\tis_paired\tmapped_is_read1\tmapped_is_read2\tis_proper_pair\tis_qcfail\tis_duplicate\n")
                for qname, index in reads.items():
                    fastq_count = &fastq_counts[index]
                    
                    # remove "\n" from base64 encoded md5
                    outfile_details.write("%s\t%s\n" % (
                        force_str(base64.encodestring(qname)[:-1]),
                        "\t".join( \
                                   map(str,
                                       (fastq_count.is_unmapped,
                                        fastq_count.mate_is_unmapped,
                                        fastq_count.is_paired,
                                        fastq_count.mapped_is_read1,
                                        fastq_count.mapped_is_read2,
                                        fastq_count.is_proper_pair,
                                        fastq_count.is_qcfail,
                                        fastq_count.is_duplicate
                                    ) )) ))

            else:
                # output to stdout much quicker
                # use puts to avoid the following error:
                # format not a string literal and no format arguments
                puts("read\tis_unmapped\tmate_is_unmapped\tis_paired\tmapped_is_read1\tmapped_is_read2\tis_proper_pair\tis_qcfail\tis_duplicate\n")
                for qname, index in reads.items():
                    fastq_count = &fastq_counts[index]
                    read_name = qname
                    printf("%s\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\n",
                           read_name,
                           fastq_count.is_unmapped,
                           fastq_count.mate_is_unmapped,
                           fastq_count.is_paired,
                           fastq_count.mapped_is_read1,
                           fastq_count.mapped_is_read2,
                           fastq_count.is_proper_pair,
                           fastq_count.is_qcfail,
                           fastq_count.is_duplicate)


    return (counter, t, 
            nh_filtered, nh_all, 
            nm_filtered, nm_all, 
            mapq_filtered, mapq_all, 
            max_hi )



'''bam2fastq.py - output fastq files from a bam-file
=================================================

:Tags: Genomics NGS Sequences BAM FASTQ Conversion

Purpose
-------

This script takes a :term:`bam` formatted file and converts to it to
one or two :term:`fastq` formatted files for single-end or paired-end
data, respectively.

For paired-end data, the first fastq file contains the first read of a
read pair and the other contains the second read of read pair.

Example
-------

For example::

   cat in.bam cgat bam2fastq out.1.fastq.gz out.2.fastq.gz

This command converts the :term:`bam` formatted file in.bam into
:term:`fastq` files containing forward reads (out.1.fastq.gz) and
reverse reads (out.2.fastq.gz).  The output files can alternatively
supplied via the option ``--output-pattern-filename``. The statement
below will create the same two output files::

   cat in.bam cgat bam2fastq --output-filename-pattern=out.%s.fastq.gz

Type::

   python bam2fastq.py --help

for command line help.

Command line options
--------------------

'''

import os
import sys
import tempfile
import shutil
import CGAT.Experiment as E
import CGAT.IOTools as IOTools

import pysam


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.set_defaults(
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    # do sth
    if len(args) == 1:
        fastqfile1 = args[0]
        fastqfile2 = options.output_filename_pattern % "2"
    elif len(args) == 2:
        fastqfile1, fastqfile2 = args
    else:
        fastqfile1 = options.output_filename_pattern % "1"
        fastqfile2 = options.output_filename_pattern % "2"

    # only output compressed data
    if not fastqfile1.endswith(".gz"):
        fastqfile1 += ".gz"
    if not fastqfile2.endswith(".gz"):
        fastqfile2 += ".gz"

    if options.stdin != sys.stdin:
        samfile = pysam.AlignmentFile(options.stdin.name, "rb")
    else:
        samfile = pysam.AlignmentFile("-", "rb")

    tmpdir = tempfile.mkdtemp()

    outtemp1 = os.path.join(tmpdir, "pair1.gz")
    outtemp2 = os.path.join(tmpdir, "pair2.gz")

    outstream1 = IOTools.openFile(outtemp1, "w")
    outstream2 = IOTools.openFile(outtemp2, "w")

    E.info('writing fastq files to temporary directory %s' % tmpdir)

    found1, found2 = set(), set()
    read1_qlen, read2_qlen = 0, 0

    c = E.Counter()
    for read in samfile.fetch(until_eof=True):
        c.input += 1
        if not read.is_paired:
            outstream1.write(
                "\t".join((read.qname, read.seq, read.qual)) + "\n")
            found1.add(read.qname)
            if not read1_qlen:
                read1_qlen = read.qlen
            c.unpaired += 1
        elif read.is_read1:
            outstream1.write(
                "\t".join((read.qname, read.seq, read.qual)) + "\n")
            found1.add(read.qname)
            if not read1_qlen:
                read1_qlen = read.qlen
            c.output1 += 1
        elif read.is_read2:
            if read.qname not in found2:
                outstream2.write(
                    "\t".join((read.qname, read.seq, read.qual)) + "\n")
                found2.add(read.qname)
                if not read2_qlen:
                    read2_qlen = read.qlen
                c.output2 += 1

    if c.unpaired == 0 and c.output1 == 0 and c.output2 == 0:
        E.warn("no reads were found")
        return

    sort_statement = '''gunzip < %s
    | sort -k1,1
    | awk '{printf("@%%s\\n%%s\\n+\\n%%s\\n", $1,$2,$3)}'
    | gzip > %s'''

    if c.output1 == 0 and c.output2 == 0:
        # single end data:
        outstream1.close()
        outstream2.close()
        E.info("sorting fastq files")
        E.run(sort_statement % (outtemp1, fastqfile1))

    else:
        # paired end data
        for qname in found2.difference(found1):
            outstream1.write(
                "\t".join((qname, "N" * read1_qlen, "B" * read1_qlen)) + "\n")
            c.extra1 += 1

        for qname in found1.difference(found2):
            outstream2.write(
                "\t".join((qname, "N" * read2_qlen, "B" * read2_qlen)) + "\n")
            c.extra2 += 1

        E.info("%s" % str(c))

        outstream1.close()
        outstream2.close()

        E.info("sorting fastq files")
        E.run(sort_statement % (outtemp1, fastqfile1))
        E.run(sort_statement % (outtemp2, fastqfile2))

    shutil.rmtree(tmpdir)

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))
